# ------------------------------------------------------------------------------
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
#   Modules
# ------------------------------------------------------------------------------

# Date
from datetime import date
from datetime import time
from datetime import datetime
from datetime import timedelta

# Filesystem
from os import W_OK
from os import mkdir
from os import access
from os import remove

from os.path import exists
from os.path import dirname
from os.path import getctime
from os.path import splitext
from os.path import basename
from os.path import expanduser
from os.path import join as path_join

from glob import glob
from shutil import copy2 as copy

# Processus
from subprocess import PIPE
from subprocess import Popen
from subprocess import STDOUT

# Random
from random import randint

# Regex
from re import match
from re import IGNORECASE

# System
from sys import exit as sys_exit
from time import sleep
from shlex import split as shlex_split
from shutil import move as rename
from platform import system

# Threading
from threading import Thread

# URL
from urllib.parse import urlparse
from urllib.request import url2pathname

# ------------------------------------------------------------------------------
#   Modules - Interface
# ------------------------------------------------------------------------------

try:
    from gi import require_version

    require_version("Gtk", "3.0")

    from gi.repository import Gtk
    from gi.repository import Gdk

    from gi.repository.Gdk import EventType

    from gi.repository.GLib import idle_add
    from gi.repository.GLib import source_remove

    from gi.repository.GObject import GObject
    from gi.repository.GObject import MainLoop
    from gi.repository.GObject import SIGNAL_RUN_LAST

    from gi.repository.GdkPixbuf import Pixbuf
    from gi.repository.GdkPixbuf import InterpType
    from gi.repository.GdkPixbuf import Colorspace

except ImportError as error:
    sys_exit("Import error with python3-gobject module: %s" % str(error))

# ------------------------------------------------------------------------------
#   Modules - GEM
# ------------------------------------------------------------------------------

try:
    from gem.utils import *
    from gem.configuration import Configuration

    from gem.api import GEM
    from gem.api import Game
    from gem.api import Console
    from gem.api import Emulator

    from gem.gtk import *
    from gem.gtk.game import GameThread
    from gem.gtk.widgets import CheckMenuItem
    from gem.gtk.widgets import ImageMenuItem
    from gem.gtk.windows import *
    from gem.gtk.mednafen import DialogMednafenMemory
    from gem.gtk.preferences import Preferences

except ImportError as error:
    sys_exit("Import error with gem module: %s" % str(error))

# ------------------------------------------------------------------------------
#   Modules - Translation
# ------------------------------------------------------------------------------

from gettext import gettext as _
from gettext import textdomain
from gettext import bindtextdomain

bindtextdomain("gem", get_data("i18n"))
textdomain("gem")

# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class Interface(Gtk.ApplicationWindow):

    __gsignals__ = { "game-terminate": (SIGNAL_RUN_LAST, None, [object]) }

    def __init__(self, api):
        """ Constructor

        Parameters
        ----------
        api : gem.api.GEM
            GEM API instance

        Raises
        ------
        TypeError
            if api type is not gem.api.GEM
        """

        if not type(api) is GEM:
            raise TypeError("Wrong type for api, expected gem.api.GEM")

        Gtk.Window.__init__(self)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        # Generate a title from GEM informations
        self.title = "%s - %s (%s)" % (GEM.Name, GEM.Version, GEM.CodeName)

        # Store thread id for game listing
        self.list_thread = int()

        # Store normal icons with icon name as key
        self.icons = dict()
        # Store alternative icons with icon name as key
        self.alternative = dict()
        # Store started notes with note file path as key
        self.notes = dict()
        # Store started threads with basename game file without extension as key
        self.threads = dict()
        # Store selected game informations with console, game and name as keys
        self.selection = dict()
        # Store shortcut with Gtk.Widget as key
        self.shortcuts_data = dict()
        # Store consoles iters
        self.consoles_iter = dict()

        # Store user keys input
        self.keys = list()
        # Store sidebar description ordre
        self.sidebar_keys = [
            ("play_time", _("Play time")),
            ("last_play", _("Last launch")),
            ("tags", _("Tags"))
        ]

        # Avoid to reload interface when switch between default & classic theme
        self.use_classic_theme = False

        # Manage fullscreen from boolean variable
        self.__fullscreen_status = False

        # Avoid to reload game tooltip every time the user move in line
        self.__current_tooltip = None
        self.__current_tooltip_data = list()
        self.__current_tooltip_pixbuf = None

        # Check mednafen status
        self.__mednafen_status = self.check_mednafen()

        # ------------------------------------
        #   Initialize API
        # ------------------------------------

        # GEM API
        self.api = api

        # Quick access to API logger
        self.logger = api.logger

        # ------------------------------------
        #   Initialize icons
        # ------------------------------------

        # Get user icon theme
        self.icons_theme = Gtk.IconTheme.get_default()

        self.icons_theme.append_search_path(get_data(path_join("icons", "ui")))

        self.icons_data = {
            "save": Icons.Floppy,
            "snap": Icons.Photos,
            "except": Icons.Properties,
            "warning": Icons.Warning,
            "favorite": Icons.Favorite,
            "multiplayer": Icons.Users,
            "finish": Icons.Smile,
            "unfinish": Icons.Uncertain
        }

        for icon in self.icons_data.keys():
            self.icons[icon] = icon_load(self.icons_data[icon], 22)

            self.alternative[icon] = None

        # HACK: Create an empty image to avoid g_object_set_qdata warning
        self.empty = Pixbuf.new(Colorspace.RGB, True, 8, 22, 22)
        self.empty.fill(0x00000000)

        # Generate symbolic icons class
        for key, value in Icons.__dict__.items():
            if not key.startswith("__") and not key.endswith("__"):
                setattr(Icons.Symbolic, key, "%s-symbolic" % value)

        # ------------------------------------
        #   Shortcuts
        # ------------------------------------

        self.shortcuts_group = Gtk.AccelGroup()

        self.shortcuts_map = Gtk.AccelMap()

        # ------------------------------------
        #   Targets
        # ------------------------------------

        self.targets = [ Gtk.TargetEntry.new("text/uri-list", 0, 1337) ]

        # ------------------------------------
        #   Toolbar size constants
        # ------------------------------------

        self.toolbar_sizes = {
            "menu": Gtk.IconSize.MENU,
            "small-toolbar": Gtk.IconSize.SMALL_TOOLBAR,
            "large-toolbar": Gtk.IconSize.LARGE_TOOLBAR,
            "button": Gtk.IconSize.BUTTON,
            "dnd": Gtk.IconSize.DND,
            "dialog": Gtk.IconSize.DIALOG }

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init packing
        self.__init_packing()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

        # ------------------------------------
        #   Main loop
        # ------------------------------------

        self.main_loop = MainLoop()
        self.main_loop.run()


    def __init_widgets(self):
        """ Initialize interface widgets
        """

        # ------------------------------------
        #   Main window
        # ------------------------------------

        self.set_title(self.title)

        self.set_icon_name(GEM.Icon)
        self.set_default_icon_name(Icons.Gaming)

        self.set_position(Gtk.WindowPosition.CENTER)
        self.set_wmclass(GEM.Acronym.upper(), GEM.Acronym.lower())

        self.add_accel_group(self.shortcuts_group)

        self.drag_dest_set(
            Gtk.DestDefaults.MOTION | Gtk.DestDefaults.DROP, self.targets,
            Gdk.DragAction.COPY)

        # ------------------------------------
        #   Clipboard
        # ------------------------------------

        self.clipboard = Gtk.Clipboard.get(Gdk.Atom.intern("CLIPBOARD", False))

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.grid = Gtk.Box()

        self.grid_games = Gtk.Box()
        self.grid_games_placeholder = Gtk.Box()
        self.grid_options = Gtk.Box()
        self.grid_infobar = Gtk.Box()

        self.grid_paned = Gtk.Box()
        self.grid_paned_widgets = Gtk.Box()
        self.grid_paned_footer = Gtk.Box()

        self.grid_toolbar_console = Gtk.Box()
        self.grid_toolbar_filters = Gtk.Box()

        self.grid_menu = Gtk.Grid()
        self.grid_menu_filters = Gtk.Grid()

        self.grid_menu_actions = Gtk.Box()
        self.grid_gem_log = Gtk.Box()

        # Properties
        self.grid.set_orientation(Gtk.Orientation.VERTICAL)

        self.grid_games.set_orientation(Gtk.Orientation.VERTICAL)

        self.grid_games_placeholder.set_spacing(12)
        self.grid_games_placeholder.set_border_width(18)
        self.grid_games_placeholder.set_orientation(Gtk.Orientation.VERTICAL)

        self.grid_paned.set_border_width(6)
        self.grid_paned.set_size_request(432, 216)
        self.grid_paned.set_orientation(Gtk.Orientation.HORIZONTAL)

        self.grid_paned_widgets.set_spacing(6)
        self.grid_paned_widgets.set_border_width(6)
        self.grid_paned_widgets.set_orientation(Gtk.Orientation.VERTICAL)

        self.grid_paned_footer.set_spacing(12)
        self.grid_paned_footer.set_margin_top(12)
        self.grid_paned_footer.set_orientation(Gtk.Orientation.HORIZONTAL)

        Gtk.StyleContext.add_class(
            self.grid_options.get_style_context(), "linked")
        self.grid_options.set_spacing(-1)
        self.grid_options.set_orientation(Gtk.Orientation.HORIZONTAL)

        Gtk.StyleContext.add_class(
            self.grid_toolbar_console.get_style_context(), "linked")
        self.grid_toolbar_console.set_spacing(-1)
        self.grid_toolbar_console.set_orientation(Gtk.Orientation.HORIZONTAL)

        Gtk.StyleContext.add_class(
            self.grid_toolbar_filters.get_style_context(), "linked")
        self.grid_toolbar_filters.set_spacing(-1)
        self.grid_toolbar_filters.set_orientation(Gtk.Orientation.HORIZONTAL)

        self.grid_menu.set_border_width(12)
        self.grid_menu.set_row_spacing(6)
        self.grid_menu.set_column_spacing(12)

        self.grid_menu_filters.set_border_width(12)
        self.grid_menu_filters.set_row_spacing(6)
        self.grid_menu_filters.set_column_spacing(12)

        Gtk.StyleContext.add_class(
            self.grid_menu_actions.get_style_context(), "linked")
        self.grid_menu_actions.set_spacing(-1)
        self.grid_menu_actions.set_orientation(Gtk.Orientation.HORIZONTAL)

        self.grid_gem_log.set_spacing(12)
        self.grid_gem_log.set_orientation(Gtk.Orientation.HORIZONTAL)

        # ------------------------------------
        #   Headerbar
        # ------------------------------------

        self.headerbar = Gtk.HeaderBar()

        self.headerbar_item_launch = Gtk.Button()

        self.headerbar_image_menu = Gtk.Image()
        self.headerbar_item_menu = Gtk.MenuButton()

        self.headerbar_image_fullscreen = Gtk.Image()
        self.headerbar_item_fullscreen = Gtk.ToggleButton()

        self.popover_menu = Gtk.Popover()

        # Properties
        self.headerbar.set_title(self.title)
        self.headerbar.set_subtitle(str())

        self.headerbar_item_launch.set_label(_("Play"))
        self.headerbar_item_launch.set_tooltip_text(
            _("Launch selected game"))
        self.headerbar_item_launch.get_style_context().add_class(
            "suggested-action")

        self.headerbar_item_menu.set_use_popover(True)

        self.headerbar_item_fullscreen.set_tooltip_text(
            _("Alternate game fullscreen mode"))

        self.popover_menu.set_modal(True)

        # ------------------------------------
        #   Headerbar - Main menu
        # ------------------------------------

        self.menu_image_preferences = Gtk.Image()
        self.menu_item_preferences = Gtk.Button()

        self.menu_image_about = Gtk.Image()
        self.menu_item_about = Gtk.Button()

        self.menu_image_quit = Gtk.Image()
        self.menu_item_quit = Gtk.Button()

        self.menu_image_gem_log = Gtk.Image()
        self.menu_label_gem_log = Gtk.Label()
        self.menu_item_gem_log = Gtk.Button()

        self.menu_label_dark_theme = Gtk.Label()
        self.menu_item_dark_theme = Gtk.Switch()

        self.menu_label_sidebar = Gtk.Label()
        self.menu_item_sidebar = Gtk.Switch()

        # Properties
        self.menu_item_preferences.set_tooltip_text(_("Preferences"))
        self.menu_item_preferences.set_image(self.menu_image_preferences)
        self.menu_item_preferences.set_use_underline(True)

        self.menu_item_about.set_tooltip_text(_("About"))
        self.menu_item_about.set_image(self.menu_image_about)
        self.menu_item_about.set_use_underline(True)

        self.menu_item_quit.set_tooltip_text(_("Quit"))
        self.menu_item_quit.set_image(self.menu_image_quit)
        self.menu_item_quit.set_use_underline(True)

        self.menu_image_gem_log.set_halign(Gtk.Align.CENTER)
        self.menu_image_gem_log.set_valign(Gtk.Align.CENTER)

        self.menu_label_gem_log.set_label(_("Output log"))
        self.menu_label_gem_log.set_halign(Gtk.Align.START)

        self.menu_item_gem_log.set_margin_top(12)
        self.menu_item_gem_log.set_relief(Gtk.ReliefStyle.NONE)

        self.menu_label_dark_theme.set_margin_top(12)
        self.menu_label_dark_theme.set_label(_("Dark theme"))
        self.menu_label_dark_theme.set_halign(Gtk.Align.END)
        self.menu_label_dark_theme.set_valign(Gtk.Align.CENTER)
        self.menu_label_dark_theme.get_style_context().add_class("dim-label")

        self.menu_item_dark_theme.set_margin_top(12)

        self.menu_label_sidebar.set_margin_top(12)
        self.menu_label_sidebar.set_label(_("Sidebar"))
        self.menu_label_sidebar.set_halign(Gtk.Align.END)
        self.menu_label_sidebar.set_valign(Gtk.Align.CENTER)
        self.menu_label_sidebar.get_style_context().add_class("dim-label")

        self.menu_item_sidebar.set_margin_top(12)

        # ------------------------------------
        #   Menubar
        # ------------------------------------

        self.menubar = Gtk.MenuBar()

        self.menubar_item_main = Gtk.MenuItem()
        self.menubar_item_edit = Gtk.MenuItem()
        self.menubar_item_tools = Gtk.MenuItem()
        self.menubar_item_help = Gtk.MenuItem()

        # Properties
        self.menubar_item_main.set_label(_("_Game"))
        self.menubar_item_main.set_use_underline(True)
        self.menubar_item_edit.set_label(_("_Edit"))
        self.menubar_item_edit.set_use_underline(True)
        self.menubar_item_tools.set_label(_("_Tools"))
        self.menubar_item_tools.set_use_underline(True)
        self.menubar_item_help.set_label(_("_Help"))
        self.menubar_item_help.set_use_underline(True)

        # ------------------------------------
        #   Menubar - Main items
        # ------------------------------------

        self.menubar_main_menu = Gtk.Menu()

        self.menubar_main_item_launch = ImageMenuItem()
        self.menubar_main_item_favorite = ImageMenuItem()
        self.menubar_main_item_multiplayer = ImageMenuItem()
        self.menubar_main_item_finish = ImageMenuItem()
        self.menubar_main_item_screenshots = ImageMenuItem()
        self.menubar_main_item_output = ImageMenuItem()
        self.menubar_main_item_notes = ImageMenuItem()
        self.menubar_main_item_quit = ImageMenuItem()

        self.menubar_main_item_fullscreen = CheckMenuItem()

        # Properties
        self.menubar_main_menu.set_reserve_toggle_size(False)

        self.menubar_main_item_launch.set_label(_("_Launch"))
        self.menubar_main_item_launch.set_image_from_icon_name(
            Icons.Symbolic.Launch)

        self.menubar_main_item_favorite.set_label(_("Mark as _favorite"))
        self.menubar_main_item_favorite.set_image_from_icon_name(
            Icons.Symbolic.Favorite)

        self.menubar_main_item_multiplayer.set_label(_("Mark as _multiplayer"))
        self.menubar_main_item_multiplayer.set_image_from_icon_name(
            Icons.Symbolic.Users)

        self.menubar_main_item_finish.set_label(_("Mark as _finish"))
        self.menubar_main_item_finish.set_image_from_icon_name(
            Icons.Symbolic.Smile)

        self.menubar_main_item_screenshots.set_label(_("Show _screenshots"))
        self.menubar_main_item_screenshots.set_image_from_icon_name(
            Icons.Symbolic.Image)

        self.menubar_main_item_output.set_label(_("Show output _log"))
        self.menubar_main_item_output.set_image_from_icon_name(
            Icons.Symbolic.Terminal)

        self.menubar_main_item_notes.set_label(_("Show _notes"))
        self.menubar_main_item_notes.set_image_from_icon_name(
            Icons.Symbolic.Document)

        self.menubar_main_item_quit.set_label(_("_Quit"))
        self.menubar_main_item_quit.set_image_from_icon_name(
            Icons.Symbolic.Quit)

        self.menubar_main_item_fullscreen.set_label(
            _("Activate game fullscreen mode"))

        # ------------------------------------
        #   Menubar - Edit items
        # ------------------------------------

        self.menubar_edit_menu = Gtk.Menu()

        self.menubar_edit_item_rename = ImageMenuItem()
        self.menubar_edit_item_parameters = ImageMenuItem()
        self.menubar_edit_item_copy = ImageMenuItem()
        self.menubar_edit_item_open = ImageMenuItem()
        self.menubar_edit_item_desktop = ImageMenuItem()
        self.menubar_edit_item_database = ImageMenuItem()
        self.menubar_edit_item_delete = ImageMenuItem()
        self.menubar_edit_item_mednafen = ImageMenuItem()

        # Properties
        self.menubar_edit_menu.set_reserve_toggle_size(False)

        self.menubar_edit_item_rename.set_label(_("_Rename"))
        self.menubar_edit_item_rename.set_image_from_icon_name(
            Icons.Symbolic.Editor)

        self.menubar_edit_item_parameters.set_label(_("Custom _parameters"))
        self.menubar_edit_item_parameters.set_image_from_icon_name(
            Icons.Symbolic.Properties)

        self.menubar_edit_item_copy.set_label(_("_Copy file path"))
        self.menubar_edit_item_copy.set_image_from_icon_name(
            Icons.Symbolic.Copy)

        self.menubar_edit_item_open.set_label(_("_Open file path"))
        self.menubar_edit_item_open.set_image_from_icon_name(
            Icons.Symbolic.Open)

        self.menubar_edit_item_desktop.set_label(_("_Generate a menu entry"))
        self.menubar_edit_item_desktop.set_image_from_icon_name(
            Icons.Symbolic.Desktop)

        self.menubar_edit_item_database.set_label(_("_Reset game informations"))
        self.menubar_edit_item_database.set_image_from_icon_name(
            Icons.Symbolic.Clear)

        self.menubar_edit_item_delete.set_label(_("_Remove game from disk"))
        self.menubar_edit_item_delete.set_image_from_icon_name(
            Icons.Symbolic.Delete)

        self.menubar_edit_item_mednafen.set_label(_("Specify a _memory type"))
        self.menubar_edit_item_mednafen.set_image_from_icon_name(
            Icons.Symbolic.Save)

        # ------------------------------------
        #   Menubar - Tools items
        # ------------------------------------

        self.menubar_tools_menu = Gtk.Menu()

        self.menubar_tools_item_preferences = ImageMenuItem()

        self.menubar_tools_item_dark_theme = CheckMenuItem()
        self.menubar_tools_item_sidebar = CheckMenuItem()

        # Properties
        self.menubar_tools_menu.set_reserve_toggle_size(False)

        self.menubar_tools_item_preferences.set_label(_("_Preferences"))
        self.menubar_tools_item_preferences.set_image_from_icon_name(
            Icons.Symbolic.System)

        self.menubar_tools_item_dark_theme.set_label(_("Use _dark theme"))
        self.menubar_tools_item_dark_theme.set_use_underline(True)

        self.menubar_tools_item_sidebar.set_label(_("Show _sidebar"))
        self.menubar_tools_item_sidebar.set_use_underline(True)

        # ------------------------------------
        #   Menubar - Help items
        # ------------------------------------

        self.menubar_help_menu = Gtk.Menu()

        self.menubar_help_item_log = ImageMenuItem()
        self.menubar_help_item_about = ImageMenuItem()

        # Properties
        self.menubar_help_menu.set_reserve_toggle_size(False)

        self.menubar_help_item_log.set_label(_("Show main _log"))
        self.menubar_help_item_log.set_image_from_icon_name(
            Icons.Symbolic.Terminal)

        self.menubar_help_item_about.set_label(_("_About"))
        self.menubar_help_item_about.set_image_from_icon_name(
            Icons.Symbolic.About)

        # ------------------------------------
        #   Toolbar
        # ------------------------------------

        self.toolbar = Gtk.Toolbar()

        self.toolbar_item_parameters = Gtk.Button()
        self.toolbar_item_screenshots = Gtk.Button()
        self.toolbar_item_output = Gtk.Button()
        self.toolbar_item_notes = Gtk.Button()
        self.toolbar_item_properties = Gtk.Button()
        self.toolbar_item_refresh = Gtk.Button()

        self.toolbar_item_game_launch = Gtk.ToolItem()
        self.toolbar_item_game_option = Gtk.ToolItem()
        self.toolbar_item_consoles = Gtk.ToolItem()
        self.toolbar_item_search = Gtk.ToolItem()
        self.toolbar_item_filters = Gtk.ToolItem()

        self.toolbar_item_separator = Gtk.SeparatorToolItem()

        self.toolbar_image_parameters = Gtk.Image()
        self.toolbar_image_screenshots = Gtk.Image()
        self.toolbar_image_output = Gtk.Image()
        self.toolbar_image_notes = Gtk.Image()
        self.toolbar_image_properties = Gtk.Image()
        self.toolbar_image_refresh = Gtk.Image()

        # Properties
        self.toolbar_item_parameters.set_tooltip_text(
            _("Set custom parameters"))

        self.toolbar_item_screenshots.set_tooltip_text(
            _("Show selected game screenshots"))
        self.toolbar_item_output.set_tooltip_text(
            _("Show selected game output log"))
        self.toolbar_item_notes.set_tooltip_text(
            _("Show selected game notes"))

        self.toolbar_item_properties.set_tooltip_text(_("Edit emulator"))
        self.toolbar_item_refresh.set_tooltip_text(_("Refresh games list"))

        self.toolbar_item_separator.set_draw(False)
        self.toolbar_item_separator.set_expand(True)

        # ------------------------------------
        #   Toolbar - Consoles
        # ------------------------------------

        self.model_consoles = Gtk.ListStore(
            Pixbuf, # Console icon
            str,    # Console name
            Pixbuf, # Emulator binary status
            str     # Console identifier
        )

        self.combo_consoles = Gtk.ComboBox()

        self.cell_consoles_icon = Gtk.CellRendererPixbuf()
        self.cell_consoles_name = Gtk.CellRendererText()
        self.cell_consoles_status = Gtk.CellRendererPixbuf()

        # Properties
        self.model_consoles.set_sort_column_id(1, Gtk.SortType.ASCENDING)

        self.combo_consoles.set_model(self.model_consoles)

        self.combo_consoles.pack_start(self.cell_consoles_icon, False)
        self.combo_consoles.pack_start(self.cell_consoles_name, True)
        self.combo_consoles.pack_start(self.cell_consoles_status, False)

        self.combo_consoles.add_attribute(
            self.cell_consoles_icon, "pixbuf", 0)
        self.combo_consoles.add_attribute(
            self.cell_consoles_name, "text", 1)
        self.combo_consoles.add_attribute(
            self.cell_consoles_status, "pixbuf", 2)

        self.cell_consoles_name.set_padding(8, 0)

        # ------------------------------------
        #   Toolbar - Filter
        # ------------------------------------

        self.entry_filter = Gtk.SearchEntry()

        self.tool_menu_filters = Gtk.MenuButton()
        self.tool_image_filters = Gtk.Image()

        self.popover_menu_filters = Gtk.Popover()

        # Properties
        self.entry_filter.set_placeholder_text(_("Filter..."))
        self.entry_filter.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, Icons.Symbolic.Clear)

        self.tool_menu_filters.set_tooltip_text(_("Filters"))
        self.tool_menu_filters.set_use_popover(True)

        self.popover_menu_filters.set_modal(True)

        # ------------------------------------
        #   Toolbar - Filter - Menu
        # ------------------------------------

        self.label_filter_favorite = Gtk.Label()
        self.check_filter_favorite = Gtk.Switch()

        self.label_filter_unfavorite = Gtk.Label()
        self.check_filter_unfavorite = Gtk.Switch()

        self.label_filter_multiplayer = Gtk.Label()
        self.check_filter_multiplayer = Gtk.Switch()

        self.label_filter_singleplayer = Gtk.Label()
        self.check_filter_singleplayer = Gtk.Switch()

        self.label_filter_finish = Gtk.Label()
        self.check_filter_finish = Gtk.Switch()

        self.label_filter_unfinish = Gtk.Label()
        self.check_filter_unfinish = Gtk.Switch()

        self.item_filter_reset = Gtk.Button()

        # Properties
        self.label_filter_favorite.set_label(_("Favorite"))
        self.label_filter_favorite.set_halign(Gtk.Align.END)
        self.label_filter_favorite.set_valign(Gtk.Align.CENTER)
        self.label_filter_favorite.get_style_context().add_class("dim-label")
        self.check_filter_favorite.set_active(True)

        self.label_filter_unfavorite.set_label(_("Unfavorite"))
        self.label_filter_unfavorite.set_halign(Gtk.Align.END)
        self.label_filter_unfavorite.set_valign(Gtk.Align.CENTER)
        self.label_filter_unfavorite.get_style_context().add_class("dim-label")
        self.check_filter_unfavorite.set_active(True)

        self.label_filter_multiplayer.set_margin_top(12)
        self.label_filter_multiplayer.set_label(_("Multiplayer"))
        self.label_filter_multiplayer.set_halign(Gtk.Align.END)
        self.label_filter_multiplayer.set_valign(Gtk.Align.CENTER)
        self.label_filter_multiplayer.get_style_context().add_class("dim-label")
        self.check_filter_multiplayer.set_margin_top(12)
        self.check_filter_multiplayer.set_active(True)

        self.label_filter_singleplayer.set_label(_("Singleplayer"))
        self.label_filter_singleplayer.set_halign(Gtk.Align.END)
        self.label_filter_singleplayer.set_valign(Gtk.Align.CENTER)
        self.label_filter_singleplayer.get_style_context().add_class(
            "dim-label")
        self.check_filter_singleplayer.set_active(True)

        self.label_filter_finish.set_margin_top(12)
        self.label_filter_finish.set_label(_("Finish"))
        self.label_filter_finish.set_halign(Gtk.Align.END)
        self.label_filter_finish.set_valign(Gtk.Align.CENTER)
        self.label_filter_finish.get_style_context().add_class("dim-label")
        self.check_filter_finish.set_margin_top(12)
        self.check_filter_finish.set_active(True)

        self.label_filter_unfinish.set_label(_("Unfinish"))
        self.label_filter_unfinish.set_halign(Gtk.Align.END)
        self.label_filter_unfinish.set_valign(Gtk.Align.CENTER)
        self.label_filter_unfinish.get_style_context().add_class("dim-label")
        self.check_filter_unfinish.set_active(True)

        self.item_filter_reset.set_margin_top(12)
        self.item_filter_reset.set_label(_("Reset filters"))
        self.item_filter_reset.set_halign(Gtk.Align.CENTER)
        self.item_filter_reset.set_valign(Gtk.Align.CENTER)

        # ------------------------------------
        #   Infobar
        # ------------------------------------

        self.infobar = Gtk.InfoBar()

        self.label_infobar = Gtk.Label()

        # Properties
        self.infobar.set_show_close_button(False)

        self.label_infobar.set_use_markup(True)

        # ------------------------------------
        #   Games - Sidebar
        # ------------------------------------

        self.paned_games = Gtk.Paned()

        self.label_game_title = Gtk.Label()
        self.label_game_description = Gtk.Label()
        self.label_game_footer = Gtk.Label()

        self.image_game_screen = Gtk.Image()

        self.separator_game = Gtk.Separator()
        self.separator_footer = Gtk.Separator()

        self.image_paned_savestates = Gtk.Image()
        self.image_paned_parameters = Gtk.Image()

        # Properties
        self.paned_games.set_orientation(Gtk.Orientation.VERTICAL)

        self.label_game_title.set_hexpand(True)
        self.label_game_title.set_alignment(0, 0.5)
        self.label_game_title.set_use_markup(True)
        self.label_game_title.set_margin_bottom(12)
        self.label_game_title.set_ellipsize(Pango.EllipsizeMode.END)

        self.label_game_description.set_use_markup(True)
        self.label_game_description.set_alignment(0, 0)
        self.label_game_description.set_ellipsize(Pango.EllipsizeMode.END)

        self.label_game_footer.set_use_markup(True)
        self.label_game_footer.set_alignment(0, 0.5)
        self.label_game_footer.set_ellipsize(Pango.EllipsizeMode.END)

        self.image_game_screen.set_alignment(0, 0)

        self.separator_game.set_no_show_all(True)
        self.separator_footer.set_no_show_all(True)

        self.image_paned_savestates.set_from_pixbuf(self.empty)
        self.image_paned_parameters.set_from_pixbuf(self.empty)

        # ------------------------------------
        #   Games - Sidebar description
        # ------------------------------------

        self.widgets_sidebar = dict()

        for key, value in self.sidebar_keys:
            self.widgets_sidebar[key] = {
                "box": Gtk.Box(),
                "key": Gtk.Label(),
                "value": Gtk.Label()
            }

            # Properties
            self.widgets_sidebar[key]["box"].set_spacing(8)
            self.widgets_sidebar[key]["box"].set_orientation(
                Gtk.Orientation.HORIZONTAL)

            self.widgets_sidebar[key]["key"].set_use_markup(True)
            self.widgets_sidebar[key]["key"].set_alignment(0, 0)
            self.widgets_sidebar[key]["key"].set_ellipsize(
                Pango.EllipsizeMode.END)
            self.widgets_sidebar[key]["key"].set_markup("<b>%s</b> :" % value)

            self.widgets_sidebar[key]["value"].set_use_markup(True)
            self.widgets_sidebar[key]["value"].set_alignment(0, 0)
            self.widgets_sidebar[key]["value"].set_ellipsize(
                Pango.EllipsizeMode.END)

            self.widgets_sidebar[key]["box"].pack_start(
                self.widgets_sidebar[key]["key"], False, False, 0)
            self.widgets_sidebar[key]["box"].pack_start(
                self.widgets_sidebar[key]["value"], True, True, 0)

        # ------------------------------------
        #   Games - Placeholder
        # ------------------------------------

        self.image_game_placeholder = Gtk.Image()
        self.label_game_placeholder = Gtk.Label()

        # Properties
        self.image_game_placeholder.set_from_icon_name(
            Icons.Symbolic.Gaming, Gtk.IconSize.DIALOG)
        self.image_game_placeholder.set_pixel_size(256)
        self.image_game_placeholder.set_halign(Gtk.Align.CENTER)
        self.image_game_placeholder.set_valign(Gtk.Align.END)
        self.image_game_placeholder.get_style_context().add_class("dim-label")

        self.label_game_placeholder.set_label(
            _("Start to play by drag & drop some ROM files into interface"))
        self.label_game_placeholder.set_halign(Gtk.Align.CENTER)
        self.label_game_placeholder.set_valign(Gtk.Align.START)

        # ------------------------------------
        #   Games - Treeview
        # ------------------------------------

        self.scroll_games = Gtk.ScrolledWindow()

        self.model_games = Gtk.ListStore(
            bool,   # Favorite status
            Pixbuf, # Favorite icon
            Pixbuf, # Multiplayer icon
            Pixbuf, # Finish icon
            str,    # Name
            str,    # Played
            str,    # Last play
            str,    # Last time play
            str,    # Time play
            str,    # Installed
            Pixbuf, # Custom parameters
            Pixbuf, # Screenshots
            Pixbuf, # Save states
            object  # Game object
        )
        self.treeview_games = Gtk.TreeView()

        self.filter_games = self.model_games.filter_new()

        self.column_game_icons = Gtk.TreeViewColumn()
        self.column_game_name = Gtk.TreeViewColumn()
        self.column_game_play = Gtk.TreeViewColumn()
        self.column_game_last_play = Gtk.TreeViewColumn()
        self.column_game_play_time = Gtk.TreeViewColumn()
        self.column_game_installed = Gtk.TreeViewColumn()
        self.column_game_flags = Gtk.TreeViewColumn()

        self.cell_game_favorite = Gtk.CellRendererPixbuf()
        self.cell_game_multiplayer = Gtk.CellRendererPixbuf()
        self.cell_game_finish = Gtk.CellRendererPixbuf()
        self.cell_game_name = Gtk.CellRendererText()
        self.cell_game_play = Gtk.CellRendererText()
        self.cell_game_last_play = Gtk.CellRendererText()
        self.cell_game_last_play_time = Gtk.CellRendererText()
        self.cell_game_play_time = Gtk.CellRendererText()
        self.cell_game_installed = Gtk.CellRendererText()
        self.cell_game_except = Gtk.CellRendererPixbuf()
        self.cell_game_snapshots = Gtk.CellRendererPixbuf()
        self.cell_game_save = Gtk.CellRendererPixbuf()

        # Properties
        self.model_games.set_sort_column_id(
            Columns.Name, Gtk.SortType.ASCENDING)

        self.treeview_games.set_model(self.filter_games)
        self.treeview_games.set_headers_clickable(False)
        self.treeview_games.set_headers_visible(True)
        self.treeview_games.set_show_expanders(False)
        self.treeview_games.set_has_tooltip(True)

        self.treeview_games.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK, self.targets, Gdk.DragAction.COPY)

        self.column_game_name.set_title(_("Name"))
        self.column_game_play.set_title(_("Launch"))
        self.column_game_last_play.set_title(_("Last launch"))
        self.column_game_play_time.set_title(_("Play time"))
        self.column_game_installed.set_title(_("Installed"))
        self.column_game_flags.set_title(_("Flags"))

        self.column_game_icons.pack_start(
            self.cell_game_favorite, False)
        self.column_game_icons.pack_start(
            self.cell_game_multiplayer, False)
        self.column_game_icons.pack_start(
            self.cell_game_finish, False)

        self.column_game_name.set_expand(True)
        self.column_game_name.set_resizable(True)
        self.column_game_name.set_min_width(100)
        self.column_game_name.set_fixed_width(300)
        self.column_game_name.pack_start(
            self.cell_game_name, True)

        self.column_game_play.set_alignment(.5)
        self.column_game_play.pack_start(
            self.cell_game_play, False)

        self.column_game_last_play.set_alignment(.5)
        self.column_game_last_play.pack_start(
            self.cell_game_last_play, False)
        self.column_game_last_play.pack_start(
            self.cell_game_last_play_time, False)

        self.column_game_play_time.set_alignment(.5)
        self.column_game_play_time.pack_start(
            self.cell_game_play_time, False)

        self.column_game_installed.set_alignment(.5)
        self.column_game_installed.pack_start(
            self.cell_game_installed, False)

        self.column_game_flags.set_alignment(.5)
        self.column_game_flags.pack_start(
            self.cell_game_except, False)
        self.column_game_flags.pack_start(
            self.cell_game_snapshots, False)
        self.column_game_flags.pack_start(
            self.cell_game_save, False)

        self.column_game_icons.add_attribute(
            self.cell_game_favorite, "pixbuf", Columns.Icon)
        self.column_game_icons.add_attribute(
            self.cell_game_multiplayer, "pixbuf", Columns.Multiplayer)
        self.column_game_icons.add_attribute(
            self.cell_game_finish, "pixbuf", Columns.Finish)
        self.column_game_name.add_attribute(
            self.cell_game_name, "text", Columns.Name)
        self.column_game_play.add_attribute(
            self.cell_game_play, "text", Columns.Played)
        self.column_game_last_play.add_attribute(
            self.cell_game_last_play, "text", Columns.LastPlay)
        self.column_game_last_play.add_attribute(
            self.cell_game_last_play_time, "text", Columns.LastTimePlay)
        self.column_game_play_time.add_attribute(
            self.cell_game_play_time, "text", Columns.TimePlay)
        self.column_game_installed.add_attribute(
            self.cell_game_installed, "text", Columns.Installed)
        self.column_game_flags.add_attribute(
            self.cell_game_except, "pixbuf", Columns.Except)
        self.column_game_flags.add_attribute(
            self.cell_game_snapshots, "pixbuf", Columns.Snapshots)
        self.column_game_flags.add_attribute(
            self.cell_game_save, "pixbuf", Columns.Save)

        self.cell_game_favorite.set_alignment(.5, .5)
        self.cell_game_multiplayer.set_alignment(.5, .5)
        self.cell_game_finish.set_alignment(.5, .5)
        self.cell_game_name.set_alignment(0, .5)
        self.cell_game_play.set_alignment(.5, .5)
        self.cell_game_last_play.set_alignment(0, .5)
        self.cell_game_last_play_time.set_alignment(1, .5)
        self.cell_game_play_time.set_alignment(.5, .5)
        self.cell_game_installed.set_alignment(.5, .5)

        self.cell_game_name.set_property("editable", True)
        self.cell_game_name.set_property("ellipsize", Pango.EllipsizeMode.END)

        self.cell_game_favorite.set_padding(4, 4)
        self.cell_game_multiplayer.set_padding(2, 4)
        self.cell_game_finish.set_padding(4, 4)
        self.cell_game_name.set_padding(6, 4)
        self.cell_game_play.set_padding(6, 4)
        self.cell_game_last_play.set_padding(6, 4)
        self.cell_game_last_play_time.set_padding(6, 4)
        self.cell_game_installed.set_padding(6, 4)
        self.cell_game_except.set_padding(4, 4)
        self.cell_game_snapshots.set_padding(2, 4)
        self.cell_game_save.set_padding(4, 4)

        # ------------------------------------
        #   Games - Menu
        # ------------------------------------

        self.menu_games = Gtk.Menu()
        self.menu_games_edit = Gtk.Menu()
        self.menu_games_tools = Gtk.Menu()

        self.menu_item_launch = ImageMenuItem()
        self.menu_item_favorite = ImageMenuItem()
        self.menu_item_multiplayer = ImageMenuItem()
        self.menu_item_finish = ImageMenuItem()
        self.menu_item_parameters = ImageMenuItem()
        self.menu_item_edit = ImageMenuItem()
        self.menu_item_tools = ImageMenuItem()

        # Properties
        self.menu_games.set_reserve_toggle_size(False)
        self.menu_games_edit.set_reserve_toggle_size(False)
        self.menu_games_tools.set_reserve_toggle_size(False)

        self.menu_item_launch.set_label(_("_Launch"))
        self.menu_item_launch.set_image_from_icon_name(
            Icons.Launch)

        self.menu_item_favorite.set_label(_("Mark as _favorite"))
        self.menu_item_favorite.set_image_from_icon_name(
            Icons.Favorite)

        self.menu_item_multiplayer.set_label(_("Mark as _multiplayer"))
        self.menu_item_multiplayer.set_image_from_icon_name(
            Icons.Users)

        self.menu_item_finish.set_label(_("Mark as _finish"))
        self.menu_item_finish.set_image_from_icon_name(
            Icons.Smile)

        self.menu_item_parameters.set_label(_("_Properties"))
        self.menu_item_parameters.set_image_from_icon_name(
            Icons.Properties)

        self.menu_item_edit.set_label(_("_Edit"))

        self.menu_item_tools.set_label(_("_Tools"))

        # ------------------------------------
        #   Games - Menu edit
        # ------------------------------------

        self.menu_item_rename = ImageMenuItem()
        self.menu_item_copy = ImageMenuItem()
        self.menu_item_open = ImageMenuItem()
        self.menu_item_remove = ImageMenuItem()
        self.menu_item_database = ImageMenuItem()

        # Properties
        self.menu_item_rename.set_label(_("_Rename"))
        self.menu_item_rename.set_image_from_icon_name(
            Icons.Editor)

        self.menu_item_copy.set_label(_("_Copy path"))
        self.menu_item_copy.set_image_from_icon_name(
            Icons.Copy)

        self.menu_item_open.set_label(_("_Open path"))
        self.menu_item_open.set_image_from_icon_name(
            Icons.Open)

        self.menu_item_database.set_label(_("_Reset game informations"))
        self.menu_item_database.set_image_from_icon_name(
            Icons.Clear)

        self.menu_item_remove.set_label(_("_Remove game from disk"))
        self.menu_item_remove.set_image_from_icon_name(
            Icons.Delete)

        # ------------------------------------
        #   Games - Menu tools
        # ------------------------------------

        self.menu_item_screenshots = ImageMenuItem()
        self.menu_item_output = ImageMenuItem()
        self.menu_item_notes = ImageMenuItem()
        self.menu_item_desktop = ImageMenuItem()
        self.menu_item_mednafen = ImageMenuItem()

        # Properties
        self.menu_item_screenshots.set_label(_("_Screenshots"))
        self.menu_item_screenshots.set_image_from_icon_name(
            Icons.Image)

        self.menu_item_output.set_label(_("Output _log"))
        self.menu_item_output.set_image_from_icon_name(
            Icons.Terminal)

        self.menu_item_notes.set_label(_("_Notes"))
        self.menu_item_notes.set_image_from_icon_name(
            Icons.Document)

        self.menu_item_desktop.set_label(_("_Generate a menu entry"))
        self.menu_item_desktop.set_image_from_icon_name(
            Icons.Desktop)

        self.menu_item_mednafen.set_label(_("Specify a _memory type"))
        self.menu_item_mednafen.set_image_from_icon_name(
            Icons.Save)

        # ------------------------------------
        #   Statusbar
        # ------------------------------------

        self.statusbar = Gtk.Statusbar()

        # Properties
        self.statusbar.set_no_show_all(True)

        self.statusbar.get_message_area().set_margin_top(0)
        self.statusbar.get_message_area().set_margin_left(0)
        self.statusbar.get_message_area().set_margin_right(0)
        self.statusbar.get_message_area().set_margin_bottom(0)


    def __init_packing(self):
        """ Initialize widgets packing in main window
        """

        # Main widgets
        self.grid.pack_start(self.menubar, False, False, 0)
        self.grid.pack_start(self.toolbar, False, False, 0)
        self.grid.pack_start(self.grid_infobar, False, False, 0)
        self.grid.pack_start(self.paned_games, True, True, 0)
        self.grid.pack_start(self.statusbar, False, False, 0)

        self.grid.set_focus_chain([
            self.headerbar,
            self.toolbar,
            self.scroll_games
        ])

        # Headerbar
        self.headerbar.pack_start(self.headerbar_item_launch)
        self.headerbar.pack_end(self.headerbar_item_menu)
        self.headerbar.pack_end(self.headerbar_item_fullscreen)

        self.toolbar_item_parameters.add(self.toolbar_image_parameters)
        self.headerbar_item_fullscreen.add(self.headerbar_image_fullscreen)
        self.headerbar_item_menu.add(self.headerbar_image_menu)

        # Headerbar menu
        self.headerbar_item_menu.set_popover(self.popover_menu)

        self.popover_menu.add(self.grid_menu)

        self.grid_menu.attach(self.grid_menu_actions, 0, 1, 2, 1)
        self.grid_menu.attach(self.menu_item_gem_log, 0, 2, 2, 1)
        self.grid_menu.attach(self.menu_label_dark_theme, 0, 3, 1, 1)
        self.grid_menu.attach(self.menu_item_dark_theme, 1, 3, 1, 1)
        self.grid_menu.attach(self.menu_label_sidebar, 0, 4, 1, 1)
        self.grid_menu.attach(self.menu_item_sidebar, 1, 4, 1, 1)

        self.grid_menu_actions.pack_start(
            self.menu_item_preferences, True, True, 0)
        self.grid_menu_actions.pack_start(
            self.menu_item_about, True, True, 0)
        self.grid_menu_actions.pack_start(
            self.menu_item_quit, True, True, 0)

        self.grid_gem_log.pack_start(self.menu_image_gem_log, False, False, 0)
        self.grid_gem_log.pack_start(self.menu_label_gem_log, True, True, 0)

        self.menu_item_gem_log.add(self.grid_gem_log)

        # Menu
        self.menubar.insert(self.menubar_item_main, -1)
        self.menubar.insert(self.menubar_item_edit, -1)
        self.menubar.insert(self.menubar_item_tools, -1)
        self.menubar.insert(self.menubar_item_help, -1)

        # Menu - Main items
        self.menubar_item_main.set_submenu(self.menubar_main_menu)

        self.menubar_main_menu.insert(self.menubar_main_item_launch, -1)
        self.menubar_main_menu.insert(Gtk.SeparatorMenuItem(), -1)
        self.menubar_main_menu.insert(self.menubar_main_item_favorite, -1)
        self.menubar_main_menu.insert(self.menubar_main_item_multiplayer, -1)
        self.menubar_main_menu.insert(self.menubar_main_item_finish, -1)
        self.menubar_main_menu.insert(Gtk.SeparatorMenuItem(), -1)
        self.menubar_main_menu.insert(self.menubar_main_item_screenshots, -1)
        self.menubar_main_menu.insert(self.menubar_main_item_output, -1)
        self.menubar_main_menu.insert(self.menubar_main_item_notes, -1)
        self.menubar_main_menu.insert(Gtk.SeparatorMenuItem(), -1)
        self.menubar_main_menu.insert(self.menubar_main_item_fullscreen, -1)
        self.menubar_main_menu.insert(Gtk.SeparatorMenuItem(), -1)
        self.menubar_main_menu.insert(self.menubar_main_item_quit, -1)

        # Menu - Edit items
        self.menubar_item_edit.set_submenu(self.menubar_edit_menu)

        self.menubar_edit_menu.insert(self.menubar_edit_item_rename, -1)
        self.menubar_edit_menu.insert(self.menubar_edit_item_parameters, -1)
        self.menubar_edit_menu.insert(self.menubar_edit_item_mednafen, -1)
        self.menubar_edit_menu.insert(Gtk.SeparatorMenuItem(), -1)
        self.menubar_edit_menu.insert(self.menubar_edit_item_copy, -1)
        self.menubar_edit_menu.insert(self.menubar_edit_item_open, -1)
        self.menubar_edit_menu.insert(self.menubar_edit_item_desktop, -1)
        self.menubar_edit_menu.insert(Gtk.SeparatorMenuItem(), -1)
        self.menubar_edit_menu.insert(self.menubar_edit_item_database, -1)
        self.menubar_edit_menu.insert(self.menubar_edit_item_delete, -1)

        # Menu - Tools items
        self.menubar_item_tools.set_submenu(self.menubar_tools_menu)

        self.menubar_tools_menu.insert(self.menubar_tools_item_dark_theme, -1)
        self.menubar_tools_menu.insert(Gtk.SeparatorMenuItem(), -1)
        self.menubar_tools_menu.insert(self.menubar_tools_item_sidebar, -1)
        self.menubar_tools_menu.insert(Gtk.SeparatorMenuItem(), -1)
        self.menubar_tools_menu.insert(self.menubar_tools_item_preferences, -1)

        # Menu - Help items
        self.menubar_item_help.set_submenu(self.menubar_help_menu)

        self.menubar_help_menu.insert(self.menubar_help_item_log, -1)
        self.menubar_help_menu.insert(Gtk.SeparatorMenuItem(), -1)
        self.menubar_help_menu.insert(self.menubar_help_item_about, -1)

        # Toolbar
        self.toolbar.insert(self.toolbar_item_game_launch, -1)
        self.toolbar.insert(Gtk.SeparatorToolItem(), -1)
        self.toolbar.insert(self.toolbar_item_game_option, -1)
        self.toolbar.insert(self.toolbar_item_separator, -1)
        self.toolbar.insert(self.toolbar_item_consoles, -1)
        self.toolbar.insert(Gtk.SeparatorToolItem(), -1)
        self.toolbar.insert(self.toolbar_item_search, -1)

        self.toolbar_item_game_launch.add(self.toolbar_item_parameters)
        self.toolbar_item_game_option.add(self.grid_options)
        self.toolbar_item_consoles.add(self.grid_toolbar_console)
        self.toolbar_item_search.add(self.grid_toolbar_filters)

        self.grid_toolbar_console.pack_start(
            self.combo_consoles, False, False, 0)
        self.grid_toolbar_console.pack_start(
            self.toolbar_item_properties, False, False, 0)
        self.grid_toolbar_console.pack_start(
            self.toolbar_item_refresh, False, False, 0)

        self.grid_toolbar_filters.pack_start(
            self.entry_filter, False, False, 0)
        self.grid_toolbar_filters.pack_start(
            self.tool_menu_filters, False, False, 0)

        self.grid_options.pack_start(
            self.toolbar_item_screenshots, False, False, 0)
        self.grid_options.pack_start(
            self.toolbar_item_output, False, False, 0)
        self.grid_options.pack_start(
            self.toolbar_item_notes, False, False, 0)

        self.tool_menu_filters.add(self.tool_image_filters)

        self.toolbar_item_screenshots.add(self.toolbar_image_screenshots)
        self.toolbar_item_output.add(self.toolbar_image_output)
        self.toolbar_item_notes.add(self.toolbar_image_notes)
        self.toolbar_item_properties.add(self.toolbar_image_properties)
        self.toolbar_item_refresh.add(self.toolbar_image_refresh)

        # Toolbar - Filters menu
        self.tool_menu_filters.set_popover(self.popover_menu_filters)

        self.grid_menu_filters.attach(
            self.label_filter_favorite, 0, 0, 1, 1)
        self.grid_menu_filters.attach(
            self.check_filter_favorite, 1, 0, 1, 1)
        self.grid_menu_filters.attach(
            self.label_filter_unfavorite, 0, 1, 1, 1)
        self.grid_menu_filters.attach(
            self.check_filter_unfavorite, 1, 1, 1, 1)
        self.grid_menu_filters.attach(
            self.label_filter_multiplayer, 0, 2, 1, 1)
        self.grid_menu_filters.attach(
            self.check_filter_multiplayer, 1, 2, 1, 1)
        self.grid_menu_filters.attach(
            self.label_filter_singleplayer, 0, 3, 1, 1)
        self.grid_menu_filters.attach(
            self.check_filter_singleplayer, 1, 3, 1, 1)
        self.grid_menu_filters.attach(
            self.label_filter_finish, 0, 4, 1, 1)
        self.grid_menu_filters.attach(
            self.check_filter_finish, 1, 4, 1, 1)
        self.grid_menu_filters.attach(
            self.label_filter_unfinish, 0, 5, 1, 1)
        self.grid_menu_filters.attach(
            self.check_filter_unfinish, 1, 5, 1, 1)
        self.grid_menu_filters.attach(
            self.item_filter_reset, 0, 6, 2, 1)

        self.popover_menu_filters.add(self.grid_menu_filters)

        # Infobar
        self.infobar.get_content_area().pack_start(
            self.label_infobar, True, True, 4)

        # Games paned
        self.paned_games.pack1(self.grid_games, True, False)
        self.paned_games.pack2(self.grid_paned, False, False)

        self.grid_paned.pack_start(self.grid_paned_widgets, True, True, 0)
        self.grid_paned.pack_start(self.image_game_screen, False, False, 0)

        self.grid_paned_widgets.pack_start(
            self.label_game_title, False, False, 0)

        for key, value in self.sidebar_keys:
            self.grid_paned_widgets.pack_start(
                self.widgets_sidebar[key]["box"], False, False, 0)

        self.grid_paned_widgets.pack_end(
            self.grid_paned_footer, False, False, 0)

        self.grid_paned_footer.pack_start(
            self.label_game_footer, True, True, 0)
        self.grid_paned_footer.pack_end(
            self.image_paned_savestates, False, False, 0)
        self.grid_paned_footer.pack_end(
            self.image_paned_parameters, False, False, 4)

        # Games
        self.grid_games.pack_start(self.grid_games_placeholder, True, True, 0)
        self.grid_games.pack_start(self.scroll_games, True, True, 0)

        # Games placeholder
        self.grid_games_placeholder.pack_start(
            self.image_game_placeholder, True, True, 0)
        self.grid_games_placeholder.pack_start(
            self.label_game_placeholder, True, True, 0)

        # Games treeview
        self.scroll_games.add(self.treeview_games)

        self.treeview_games.append_column(self.column_game_icons)
        self.treeview_games.append_column(self.column_game_name)
        self.treeview_games.append_column(self.column_game_play)
        self.treeview_games.append_column(self.column_game_last_play)
        self.treeview_games.append_column(self.column_game_play_time)
        self.treeview_games.append_column(self.column_game_installed)
        self.treeview_games.append_column(self.column_game_flags)

        # Games menu
        self.menu_games.append(self.menu_item_launch)
        self.menu_games.append(Gtk.SeparatorMenuItem())
        self.menu_games.append(self.menu_item_favorite)
        self.menu_games.append(self.menu_item_multiplayer)
        self.menu_games.append(self.menu_item_finish)
        self.menu_games.append(Gtk.SeparatorMenuItem())
        self.menu_games.append(self.menu_item_parameters)
        self.menu_games.append(Gtk.SeparatorMenuItem())
        self.menu_games.append(self.menu_item_edit)
        self.menu_games.append(self.menu_item_tools)

        self.menu_item_edit.set_submenu(self.menu_games_edit)
        self.menu_item_tools.set_submenu(self.menu_games_tools)

        self.menu_games_edit.append(self.menu_item_rename)
        self.menu_games_edit.append(Gtk.SeparatorMenuItem())
        self.menu_games_edit.append(self.menu_item_copy)
        self.menu_games_edit.append(self.menu_item_open)
        self.menu_games_edit.append(Gtk.SeparatorMenuItem())
        self.menu_games_edit.append(self.menu_item_remove)
        self.menu_games_edit.append(self.menu_item_database)

        self.menu_games_tools.append(self.menu_item_screenshots)
        self.menu_games_tools.append(self.menu_item_output)
        self.menu_games_tools.append(self.menu_item_notes)
        self.menu_games_tools.append(Gtk.SeparatorMenuItem())
        self.menu_games_tools.append(self.menu_item_desktop)
        self.menu_games_tools.append(Gtk.SeparatorMenuItem())
        self.menu_games_tools.append(self.menu_item_mednafen)

        self.add(self.grid)


    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.connect("game-terminate", self.__on_game_terminate)

        # ------------------------------------
        #   Window
        # ------------------------------------

        self.connect(
            "delete-event", self.__stop_interface)
        self.connect(
            "key-press-event", self.__on_manage_keys)

        self.connect(
            "drag-data-received", self.__on_dnd_received_data)

        # ------------------------------------
        #   Menubar
        # ------------------------------------

        self.menubar_main_item_launch.connect(
            "activate", self.__on_game_launch)
        self.menubar_main_item_favorite.connect(
            "activate", self.__on_game_marked_as_favorite)
        self.menubar_main_item_multiplayer.connect(
            "activate", self.__on_game_marked_as_multiplayer)
        self.menubar_main_item_finish.connect(
            "activate", self.__on_game_marked_as_finish)
        self.menubar_main_item_screenshots.connect(
            "activate", self.__on_show_viewer)
        self.menubar_main_item_output.connect(
            "activate", self.__on_show_log)
        self.menubar_main_item_notes.connect(
            "activate", self.__on_show_notes)
        self.fullscreen_signal = self.menubar_main_item_fullscreen.connect(
            "toggled", self.__on_activate_fullscreen)
        self.menubar_main_item_fullscreen.connect(
            "activate", self.__on_activate_fullscreen)
        self.menubar_main_item_quit.connect(
            "activate", self.__stop_interface)

        self.menubar_edit_item_rename.connect(
            "activate", self.__on_activate_renamed)
        self.menubar_edit_item_parameters.connect(
            "activate", self.__on_game_parameters)
        self.menubar_edit_item_copy.connect(
            "activate", self.__on_game_copy)
        self.menubar_edit_item_open.connect(
            "activate", self.__on_game_open)
        self.menubar_edit_item_desktop.connect(
            "activate", self.__on_game_generate_desktop)
        self.menubar_edit_item_database.connect(
            "activate", self.__on_game_clean)
        self.menubar_edit_item_delete.connect(
            "activate", self.__on_game_removed)
        self.menubar_edit_item_mednafen.connect(
            "activate", self.__on_game_backup_memory)

        self.menubar_tools_item_preferences.connect(
            "activate", self.__on_show_preferences)
        self.dark_signal_menubar = self.menubar_tools_item_dark_theme.connect(
            "toggled", self.__on_activate_dark_theme)
        self.menubar_tools_item_dark_theme.connect(
            "activate", self.__on_activate_dark_theme)
        self.side_signal_menubar = self.menubar_tools_item_sidebar.connect(
            "toggled", self.__on_activate_sidebar)
        self.menubar_tools_item_sidebar.connect(
            "activate", self.__on_activate_sidebar)

        self.menubar_help_item_log.connect(
            "activate", self.__on_show_log)
        self.menubar_help_item_about.connect(
            "activate", self.__on_show_about)

        # ------------------------------------
        #   Toolbar
        # ------------------------------------

        self.headerbar_item_launch.connect(
            "clicked", self.__on_game_launch)
        self.fullscreen_signal_tool = self.headerbar_item_fullscreen.connect(
            "toggled", self.__on_activate_fullscreen)
        self.toolbar_item_screenshots.connect(
            "clicked", self.__on_show_viewer)
        self.toolbar_item_output.connect(
            "clicked", self.__on_game_log)
        self.toolbar_item_notes.connect(
            "clicked", self.__on_show_notes)
        self.toolbar_item_parameters.connect(
            "clicked", self.__on_game_parameters)

        self.toolbar_item_properties.connect(
            "clicked", self.__on_show_emulator_config)
        self.toolbar_item_refresh.connect(
            "clicked", self.__on_selected_console)

        self.entry_filter.connect(
            "icon-press", on_entry_clear)
        self.entry_filter.connect(
            "changed", self.filters_update)

        self.check_filter_favorite.connect(
            "state-set", self.filters_update)
        self.check_filter_unfavorite.connect(
            "state-set", self.filters_update)
        self.check_filter_multiplayer.connect(
            "state-set", self.filters_update)
        self.check_filter_singleplayer.connect(
            "state-set", self.filters_update)
        self.check_filter_finish.connect(
            "state-set", self.filters_update)
        self.check_filter_unfinish.connect(
            "state-set", self.filters_update)
        self.item_filter_reset.connect(
            "clicked", self.filters_reset)

        # ------------------------------------
        #   Menu
        # ------------------------------------

        self.menu_item_launch.connect(
            "activate", self.__on_game_launch)
        self.menu_item_favorite.connect(
            "activate", self.__on_game_marked_as_favorite)
        self.menu_item_multiplayer.connect(
            "activate", self.__on_game_marked_as_multiplayer)
        self.menu_item_finish.connect(
            "activate", self.__on_game_marked_as_finish)
        self.menu_item_screenshots.connect(
            "activate", self.__on_show_viewer)
        self.menu_item_output.connect(
            "activate", self.__on_game_log)
        self.menu_item_notes.connect(
            "activate", self.__on_show_notes)

        self.menu_item_rename.connect(
            "activate", self.__on_activate_renamed)
        self.menu_item_parameters.connect(
            "activate", self.__on_game_parameters)
        self.menu_item_copy.connect(
            "activate", self.__on_game_copy)
        self.menu_item_open.connect(
            "activate", self.__on_game_open)
        self.menu_item_desktop.connect(
            "activate", self.__on_game_generate_desktop)
        self.menu_item_database.connect(
            "activate", self.__on_game_clean)
        self.menu_item_remove.connect(
            "activate", self.__on_game_removed)
        self.menu_item_mednafen.connect(
            "activate", self.__on_game_backup_memory)

        self.menu_item_preferences.connect(
            "clicked", self.__on_show_preferences)
        self.menu_item_gem_log.connect(
            "clicked", self.__on_show_log)
        self.dark_signal_menu = self.menu_item_dark_theme.connect(
            "state-set", self.__on_activate_dark_theme)
        self.side_signal_menu = self.menu_item_sidebar.connect(
            "state-set", self.__on_activate_sidebar)
        self.menu_item_about.connect(
            "clicked", self.__on_show_about)
        self.menu_item_quit.connect(
            "clicked", self.__stop_interface)

        # ------------------------------------
        #   Consoles
        # ------------------------------------

        self.combo_consoles.connect(
            "changed", self.__on_selected_console)

        # ------------------------------------
        #   Games
        # ------------------------------------

        self.cell_game_name.connect(
            "edited", self.__on_game_renamed)
        self.cell_game_name.connect(
            "editing-started", self.__on_game_renamed_started)
        self.cell_game_name.connect(
            "editing-canceled", self.__on_game_renamed_cancel)

        self.treeview_games.connect(
            "button-press-event", self.__on_selected_game)
        self.treeview_games.connect(
            "key-release-event", self.__on_selected_game)

        self.treeview_games.connect(
            "button-press-event", self.__on_menu_show)
        self.treeview_games.connect(
            "key-release-event", self.__on_menu_show)

        self.treeview_games.connect(
            "drag-data-get", self.__on_dnd_send_data)

        self.treeview_games.connect(
            "query-tooltip", self.__on_selected_game_tooltip)

        self.filter_games.set_visible_func(self.filters_match)


    def __start_interface(self):
        """ Load data and start interface
        """

        self.load_interface(True)

        load_console_startup = \
            self.config.getboolean("gem", "load_console_startup", fallback=True)

        # Check last loaded console in gem.conf
        if load_console_startup:
            console = self.config.item("gem", "last_console", str())

            # A console has been saved
            if len(console) > 0:

                # Check if this console use the old console name value (< 0.8)
                if not console in self.api.consoles.keys():
                    console = generate_identifier(console)

                # Check if current identifier exists
                if console in self.api.consoles.keys() and \
                    console in self.consoles_iter.keys():
                    self.treeview_games.set_visible(True)

                    # Set console combobox active iter
                    self.combo_consoles.set_active_iter(
                        self.consoles_iter[console])

                    # Set Console object as selected
                    self.selection["console"] = self.api.get_console(console)

        # Check welcome message status in gem.conf
        if self.config.getboolean("gem", "welcome", fallback=True):

            # Load the first console to avoid mini combobox
            if load_console_startup and self.selection["console"] is None:
                self.combo_consoles.set_active(0)

            dialog = Message(self, _("Welcome !"), _("Welcome and thanks for "
                "choosing GEM as emulators manager. Start using GEM by "
                "droping some roms into interface.\n\nEnjoy and have fun :D"),
                Icons.SmileBig, False)

            dialog.set_size(500, -1)

            dialog.run()
            dialog.destroy()

            # Disable welcome message for next launch
            self.config.modify("gem", "welcome", 0)
            self.config.update()

        # Set default filters flag
        self.filters_reset()


    def __stop_interface(self, *args):
        """ Save data and stop interface
        """

        # ------------------------------------
        #   Threads
        # ------------------------------------

        # Remove games listing thread
        if not self.list_thread == 0:
            source_remove(self.list_thread)

        # Remove games launcher thread
        if len(self.threads.keys()) > 0:
            for thread in self.threads.copy().keys():
                self.threads[thread].proc.terminate()

        # ------------------------------------
        #   Notes
        # ------------------------------------

        # Close open notes dialog
        if len(self.notes.keys()) > 0:
            for dialog in self.notes.copy().keys():
                self.notes[dialog].response(Gtk.ResponseType.APPLY)

        # ------------------------------------
        #   Last console
        # ------------------------------------

        # Save current console as last_console in gem.conf
        row = self.combo_consoles.get_active_iter()
        if row is not None:
            console = self.model_consoles.get_value(row, 3)

            last_console = self.config.item("gem", "last_console", None)

            # Avoid to modify gem.conf if console is already in conf
            if last_console is None or not last_console == console:
                self.config.modify("gem", "last_console", console)
                self.config.update()

                self.logger.info(
                    _("Save %s console for next startup") % console)

        # ------------------------------------
        #   Windows size
        # ------------------------------------

        self.config.modify("windows", "main", "%dx%d" % self.get_size())
        self.config.update()

        self.logger.info(_("Close interface"))

        self.main_loop.quit()


    def load_interface(self, init_interface=False):
        """ Load main interface

        Others Parameters
        -----------------
        init_interface : bool
            Interface first initialization (Default: False)
        """

        self.api.init()

        # ------------------------------------
        #   Configuration
        # ------------------------------------

        self.config = Configuration(path_join(GEM.Config, "gem.conf"))

        # Get missing keys from config/gem.conf
        self.config.add_missing_data(get_data(path_join("config", "gem.conf")))

        # ------------------------------------
        #   Shortcuts
        # ------------------------------------

        self.__init_shortcuts()

        # ------------------------------------
        #   Theme
        # ------------------------------------

        # Block signal to avoid stack overflow when toggled
        self.menu_item_dark_theme.handler_block(
            self.dark_signal_menu)
        self.menubar_tools_item_dark_theme.handler_block(
            self.dark_signal_menubar)

        dark_theme_status = self.config.getboolean(
            "gem", "dark_theme", fallback=False)

        on_change_theme(dark_theme_status)

        self.menu_item_dark_theme.set_active(dark_theme_status)
        self.menubar_tools_item_dark_theme.set_active(dark_theme_status)

        # Unblock signal
        self.menu_item_dark_theme.handler_unblock(
            self.dark_signal_menu)
        self.menubar_tools_item_dark_theme.handler_unblock(
            self.dark_signal_menubar)

        if dark_theme_status:
            self.logger.debug("Use dark variant for GTK+ theme")
        else:
            self.logger.debug("Use light variant for GTK+ theme")

        # ------------------------------------
        #   Toolbar
        # ------------------------------------

        icon_size = self.config.get(
            "gem", "toolbar_icons_size", fallback="small-toolbar")

        if init_interface or \
            not self.toolbar_sizes[icon_size] == self.toolbar.get_icon_size():
            self.headerbar_image_menu.set_from_icon_name(
                Icons.Symbolic.Menu, self.toolbar_sizes[icon_size])
            self.headerbar_image_fullscreen.set_from_icon_name(
                Icons.Symbolic.Restore, self.toolbar_sizes[icon_size])

            self.menu_image_preferences.set_from_icon_name(
                Icons.Symbolic.System, self.toolbar_sizes[icon_size])
            self.menu_image_about.set_from_icon_name(
                Icons.Symbolic.About, self.toolbar_sizes[icon_size])
            self.menu_image_quit.set_from_icon_name(
                Icons.Symbolic.Quit, self.toolbar_sizes[icon_size])
            self.menu_image_gem_log.set_from_icon_name(
                Icons.Symbolic.Terminal, self.toolbar_sizes[icon_size])

            self.toolbar.set_icon_size(self.toolbar_sizes[icon_size])

            self.toolbar_image_parameters.set_from_icon_name(
                Icons.Symbolic.Gaming, self.toolbar.get_icon_size())
            self.toolbar_image_screenshots.set_from_icon_name(
                Icons.Symbolic.Camera, self.toolbar.get_icon_size())
            self.toolbar_image_output.set_from_icon_name(
                Icons.Symbolic.Terminal, self.toolbar.get_icon_size())
            self.toolbar_image_notes.set_from_icon_name(
                Icons.Symbolic.Editor, self.toolbar.get_icon_size())
            self.toolbar_image_properties.set_from_icon_name(
                Icons.Symbolic.Properties, self.toolbar.get_icon_size())
            self.toolbar_image_refresh.set_from_icon_name(
                Icons.Symbolic.Refresh, self.toolbar.get_icon_size())

            self.tool_image_filters.set_from_icon_name(
                Icons.Symbolic.Sync, self.toolbar.get_icon_size())

        # ------------------------------------
        #   Icons
        # ------------------------------------

        icon = self.alternative["save"]

        status = self.config.getboolean(
            "gem", "use_translucent_icons", fallback=False)

        if icon is None or (icon == self.empty and status) or (
            not icon == self.empty and not status):

            for icon in self.icons_data.keys():
                if status:
                    self.alternative[icon] = set_pixbuf_opacity(
                        self.icons[icon], 50)

                else:
                    self.alternative[icon] = self.empty

        # ------------------------------------
        #   Window first drawing
        # ------------------------------------

        if init_interface:

            self.use_classic_theme = self.config.getboolean(
                "gem", "use_classic_theme", fallback=False)

            # ------------------------------------
            #   Window classic theme
            # ------------------------------------

            if not self.use_classic_theme:
                self.set_titlebar(self.headerbar)

            # ------------------------------------
            #   Window size
            # ------------------------------------

            try:
                width, height = self.config.get(
                    "windows", "main", fallback="1024x768").split('x')

                self.set_default_size(int(width), int(height))
                self.resize(int(width), int(height))

            except ValueError as error:
                self.logger.error(
                    _("Cannot resize main window: %s") % str(error))

                self.set_default_size(1024, 768)

            self.set_position(Gtk.WindowPosition.CENTER)

            self.hide()
            self.unrealize()

        # ------------------------------------
        #   Widgets
        # ------------------------------------

        self.show_all()

        self.menu_games.show_all()

        self.grid_menu.show_all()
        self.grid_menu_filters.show_all()

        self.infobar.show_all()
        self.infobar.get_content_area().show_all()

        if self.use_classic_theme:
            self.logger.debug("Use classic theme for GTK+ interface")
            self.menubar.show_all()
            self.statusbar.show()

        else:
            self.logger.debug("Use default theme for GTK+ interface")
            self.menubar.hide()
            self.statusbar.hide()

        self.set_infobar()
        self.sensitive_interface()

        # ------------------------------------
        #   Header
        # ------------------------------------

        if not self.config.getboolean("gem", "show_header", fallback=True):
            self.headerbar.set_show_close_button(False)
        else:
            self.headerbar.set_show_close_button(True)

        # ------------------------------------
        #   Sidebar
        # ------------------------------------

        sidebar_status = self.config.getboolean(
            "gem", "show_sidebar", fallback=True)

        # Block signal to avoid stack overflow when toggled
        self.menu_item_sidebar.handler_block(
            self.side_signal_menu)
        self.menubar_tools_item_sidebar.handler_block(
            self.side_signal_menubar)

        self.menu_item_sidebar.set_active(sidebar_status)
        self.menubar_tools_item_sidebar.set_active(sidebar_status)

        # Unblock signal
        self.menu_item_sidebar.handler_unblock(
            self.side_signal_menu)
        self.menubar_tools_item_sidebar.handler_unblock(
            self.side_signal_menubar)

        # Avoid to reload paned_game if user has not change orientation
        previous_mode = self.paned_games.get_orientation()

        # Wanted sidebar orientation
        sidebar_orientation = self.config.get("gem", "sidebar_orientation")

        if sidebar_orientation == "horizontal" and \
            not previous_mode == Gtk.Orientation.HORIZONTAL:
            self.paned_games.set_position(-1)
            self.paned_games.set_orientation(Gtk.Orientation.HORIZONTAL)

            self.grid_paned.set_orientation(Gtk.Orientation.VERTICAL)
            self.grid_paned.reorder_child(self.image_game_screen, 0)

            self.image_game_screen.set_alignment(0.5, 0)
            self.label_game_title.set_alignment(0.5, 0)

        elif sidebar_orientation == "vertical" and \
            not previous_mode == Gtk.Orientation.VERTICAL:
            self.paned_games.set_position(-1)
            self.paned_games.set_orientation(Gtk.Orientation.VERTICAL)

            self.grid_paned.set_orientation(Gtk.Orientation.HORIZONTAL)
            self.grid_paned.reorder_child(self.image_game_screen, -1)

            self.image_game_screen.set_alignment(0, 0)
            self.label_game_title.set_alignment(0, 0)

        if sidebar_status:
            self.grid_paned.show_all()
        else:
            self.grid_paned.hide()

        for key, value in self.sidebar_keys:
            self.widgets_sidebar[key]["box"].hide()

        # ------------------------------------
        #   Games
        # ------------------------------------

        # Game rename status
        self.is_rename = False

        # Games - Treeview
        lines = {
            "none": Gtk.TreeViewGridLines.NONE,
            "horizontal": Gtk.TreeViewGridLines.HORIZONTAL,
            "vertical": Gtk.TreeViewGridLines.VERTICAL,
            "both": Gtk.TreeViewGridLines.BOTH }

        if self.config.item("gem", "games_treeview_lines", "none") in lines:
            self.treeview_games.set_grid_lines(
                lines[self.config.item("gem", "games_treeview_lines", "none")])

        # Games - Treeview columns
        columns = {
            "play": self.column_game_play,
            "last_play": self.column_game_last_play,
            "play_time": self.column_game_play_time,
            "installed": self.column_game_installed,
            "flags": self.column_game_flags }

        for key, widget in columns.items():
            if not self.config.getboolean("columns", key, fallback=True):
                widget.set_visible(False)
            else:
                widget.set_visible(True)

        # ------------------------------------
        #   Console
        # ------------------------------------

        self.scroll_games.set_visible(False)
        self.grid_games_placeholder.set_visible(False)

        current_console = self.append_consoles()

        # Show games placeholder when no console available or selected
        if current_console is None or len(self.model_consoles) == 0:
            self.grid_games_placeholder.set_visible(True)

        self.selection = dict(
            console=None,
            game=None)

        if current_console is None:
            self.model_games.clear()

            if len(self.model_consoles) > 0:
                self.combo_consoles.set_active(0)

        else:
            self.combo_consoles.set_active_iter(current_console)


    def sensitive_interface(self, status=False):
        """ Update sensitive status for main widgets

        Parameters
        ----------
        status : bool
            Sensitive status
        """

        self.__on_game_launch_button_update(True)
        self.headerbar_item_launch.set_sensitive(status)

        self.menu_item_rename.set_sensitive(status)
        self.menu_item_favorite.set_sensitive(status)
        self.menu_item_multiplayer.set_sensitive(status)
        self.menu_item_finish.set_sensitive(status)
        self.menu_item_screenshots.set_sensitive(status)
        self.menu_item_output.set_sensitive(status)
        self.menu_item_notes.set_sensitive(status)
        self.menu_item_copy.set_sensitive(status)
        self.menu_item_open.set_sensitive(status)
        self.menu_item_desktop.set_sensitive(status)
        self.menu_item_remove.set_sensitive(status)
        self.menu_item_database.set_sensitive(status)
        self.menu_item_mednafen.set_sensitive(status)

        self.toolbar_item_output.set_sensitive(status)
        self.toolbar_item_notes.set_sensitive(status)
        self.toolbar_item_parameters.set_sensitive(status)
        self.toolbar_item_screenshots.set_sensitive(status)
        self.toolbar_item_properties.set_sensitive(status)
        self.toolbar_item_refresh.set_sensitive(status)

        self.menubar_main_item_launch.set_sensitive(status)
        self.menubar_main_item_favorite.set_sensitive(status)
        self.menubar_main_item_multiplayer.set_sensitive(status)
        self.menubar_main_item_finish.set_sensitive(status)
        self.menubar_main_item_screenshots.set_sensitive(status)
        self.menubar_main_item_output.set_sensitive(status)
        self.menubar_main_item_notes.set_sensitive(status)
        self.menubar_edit_item_rename.set_sensitive(status)
        self.menubar_edit_item_parameters.set_sensitive(status)
        self.menubar_edit_item_copy.set_sensitive(status)
        self.menubar_edit_item_open.set_sensitive(status)
        self.menubar_edit_item_desktop.set_sensitive(status)
        self.menubar_edit_item_database.set_sensitive(status)
        self.menubar_edit_item_delete.set_sensitive(status)
        self.menubar_edit_item_mednafen.set_sensitive(status)


    def filters_update(self, widget, status=None):
        """ Reload packages filter when user change filters from menu

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal

        Other Parameters
        ----------------
        status : bool or None
            New status for current widget

        Notes
        -----
        Check widget utility in this function
        """

        if status is not None and type(status) is bool:
            widget.set_active(status)

        self.filter_games.refilter()

        self.check_selection()

        self.set_informations_headerbar(self.selection["game"])


    def filters_reset(self, widget=None, events=None):
        """ Reset game filters
        """

        for child in self.grid_menu_filters.get_children():

            if type(child) is Gtk.Switch:
                child.set_active(True)


    def filters_match(self, model, row, data=None):
        """ Update treeview rows

        This function update games treeview with filter entry content. A row is
        visible if the content match the filter.

        Parameters
        ----------
        model : Gtk.TreeModel
            Treeview model which receive signal
        row : Gtk.TreeModelRow
            Treeview current row

        Other Parameters
        ----------------
        data : object
            User data to pass to the visible function (Default: None)
        """

        found = False

        # Get game object from treeview
        game = model.get_value(row, Columns.Object)

        try:
            text = self.entry_filter.get_text()

            # No filter
            if len(text) == 0:
                found = True

            # ------------------------------------
            #   Check filter
            # ------------------------------------

            # Check game name first
            if game.name is not None:

                # Regex match game.name
                if match("%s$" % text, game.name) is not None:
                    found = True

                # Lowercase filter match lowercase game.name
                if text.lower() in game.name.lower():
                    found = True

            # Check game tags second
            if len(game.tags) > 0:

                for tag in game.tags:

                    # Regex match one of game tag
                    if match("%s$" % text, tag) is not None:
                        found = True

                    # Lowercase filter match lowercase game.name
                    if text.lower() in tag.lower():
                        found = True

            # ------------------------------------
            #   Set status
            # ------------------------------------

            flags = [
                (
                    self.check_filter_favorite.get_active(),
                    self.check_filter_unfavorite.get_active(),
                    game.favorite
                ),
                (
                    self.check_filter_multiplayer.get_active(),
                    self.check_filter_singleplayer.get_active(),
                    game.multiplayer
                ),
                (
                    self.check_filter_finish.get_active(),
                    self.check_filter_unfinish.get_active(),
                    game.finish
                )
            ]

            for first, second, status in flags:

                # Check if one the two checkbox is not active
                if not (first and second):
                    found = found and (
                        (status and first) or (not status and second))

        except:
            pass

        return found


    def __init_shortcuts(self):
        """ Generate shortcuts signals from user configuration
        """

        shortcuts = [
            {
                "path": "<GEM>/game/open",
                "widgets": [
                    self.menu_item_open,
                    self.menubar_edit_item_open
                ],
                "keys": self.config.item("keys", "open", "<Control>O"),
                "function": self.__on_game_open
            },
            {
                "path": "<GEM>/game/copy",
                "widgets": [
                    self.menu_item_copy,
                    self.menubar_edit_item_copy
                ],
                "keys": self.config.item("keys", "copy", "<Control>C"),
                "function": self.__on_game_copy
            },
            {
                "path": "<GEM>/game/desktop",
                "widgets": [
                    self.menu_item_desktop,
                    self.menubar_edit_item_desktop
                ],
                "keys": self.config.item("keys", "desktop", "<Control>G"),
                "function": self.__on_game_generate_desktop
            },
            {
                "path": "<GEM>/game/start",
                "widgets": [
                    self.menu_item_launch,
                    self.menubar_main_item_launch
                ],
                "keys": self.config.item("keys", "start", "Return"),
                "function": self.__on_game_launch
            },
            {
                "path": "<GEM>/game/rename",
                "widgets": [
                    self.menu_item_rename,
                    self.menubar_edit_item_rename
                ],
                "keys": self.config.item("keys", "rename", "F2"),
                "function": self.__on_activate_renamed
            },
            {
                "path": "<GEM>/game/favorite",
                "widgets": [
                    self.menu_item_favorite,
                    self.menubar_main_item_favorite
                ],
                "keys": self.config.item("keys", "favorite", "F3"),
                "function": self.__on_game_marked_as_favorite
            },
            {
                "path": "<GEM>/game/multiplayer",
                "widgets": [
                    self.menu_item_multiplayer,
                    self.menubar_main_item_multiplayer
                ],
                "keys": self.config.item("keys", "multiplayer", "F4"),
                "function": self.__on_game_marked_as_multiplayer
            },
            {
                "path": "<GEM>/game/finish",
                "widgets": [
                    self.menu_item_finish,
                    self.menubar_main_item_finish
                ],
                "keys": self.config.item("keys", "finish", "<Control>F3"),
                "function": self.__on_game_marked_as_finish
            },
            {
                "path": "<GEM>/game/screenshots",
                "widgets": [
                    self.menu_item_screenshots,
                    self.menubar_main_item_screenshots,
                    self.toolbar_item_screenshots
                ],
                "keys": self.config.item("keys", "snapshots", "F5"),
                "function": self.__on_show_viewer
            },
            {
                "path": "<GEM>/game/log",
                "widgets": [
                    self.menu_item_output,
                    self.menubar_main_item_output
                ],
                "keys": self.config.item("keys", "log", "F6"),
                "function": self.__on_game_log
            },
            {
                "path": "<GEM>/game/notes",
                "widgets": [
                    self.menu_item_notes,
                    self.menubar_main_item_notes
                ],
                "keys": self.config.item("keys", "notes", "F7"),
                "function": self.__on_show_notes
            },
            {
                "path": "<GEM>/game/memory",
                "widgets": [
                    self.menu_item_mednafen,
                    self.menubar_edit_item_mednafen
                ],
                "keys": self.config.item("keys", "memory", "F8"),
                "function": self.__on_game_backup_memory
            },
            {
                "path": "<GEM>/sidebar",
                "widgets": [
                    self.menu_item_sidebar,
                    self.menubar_tools_item_sidebar
                ],
                "keys": self.config.item("keys", "sidebar", "F9"),
                "function": self.__on_activate_sidebar
            },
            {
                "path": "<GEM>/game/exceptions",
                "widgets": [
                    self.menu_item_parameters,
                    self.menubar_edit_item_parameters
                ],
                "keys": self.config.item("keys", "exceptions", "F12"),
                "function": self.__on_game_parameters
            },
            {
                "path": "<GEM>/game/delete",
                "widgets": [
                    self.menu_item_remove,
                    self.menubar_edit_item_delete
                ],
                "keys": self.config.item("keys", "delete", "<Control>Delete"),
                "function": self.__on_game_removed
            },
            {
                "path": "<GEM>/game/remove",
                "widgets": [
                    self.menu_item_database,
                    self.menubar_edit_item_database
                ],
                "keys": self.config.item("keys", "remove", "Delete"),
                "function": self.__on_game_clean
            },
            {
                "path": "<GEM>/preferences",
                "widgets": [
                    self.menu_item_preferences,
                    self.menubar_tools_item_preferences
                ],
                "keys": self.config.item("keys", "preferences", "<Control>P"),
                "function": self.__on_show_preferences
            },
            {
                "path": "<GEM>/log",
                "widgets": [
                    self.menu_item_gem_log,
                    self.menubar_help_item_log
                ],
                "keys": self.config.item("keys", "gem", "<Control>L"),
                "function": self.__on_show_log
            },
            {
                "path": "<GEM>/quit",
                "widgets": [
                    self.menu_item_quit,
                    self.menubar_main_item_quit
                ],
                "keys": self.config.item("keys", "quit", "<Control>Q"),
                "function": self.__stop_interface
            }
        ]

        # HACK: Get the maximum characters width of every shortcuts
        max_width = int()
        for data in shortcuts:
            key, mod = Gtk.accelerator_parse(data["keys"])
            accel = len(Gtk.accelerator_get_label(key, mod))

            if accel > max_width:
                max_width = accel

        for data in shortcuts:
            key, mod = Gtk.accelerator_parse(data["keys"])

            if Gtk.accelerator_valid(key, mod):
                # Disconnect previous shortcut to avoid multiple allocation
                self.shortcuts_group.disconnect_key(key, mod)

                self.shortcuts_group.connect(
                    key, mod, Gtk.AccelFlags.VISIBLE, data["function"])

                self.shortcuts_map.change_entry(data["path"], key, mod, True)

                for widget in data["widgets"]:
                    if type(widget) == ImageMenuItem:
                        widget.set_accelerator(max_width, key, mod)


    def __on_manage_keys(self, widget, event):
        """ Manage widgets for specific keymaps

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        event : Gdk.EventButton or Gdk.EventKey
            Event which triggered this signal
        """

        # Fullscreen switch
        if event.keyval == Gdk.KEY_F11:
            self.headerbar_item_fullscreen.set_active(
                not self.headerbar_item_fullscreen.get_active())

        # Give me more lifes, powerups or cookies konami code, I need more
        konami_code = [Gdk.KEY_Up, Gdk.KEY_Up, Gdk.KEY_Down, Gdk.KEY_Down,
            Gdk.KEY_Left, Gdk.KEY_Right, Gdk.KEY_Left, Gdk.KEY_Right]

        if event.keyval in konami_code:
            self.keys.append(event.keyval)

            if self.keys == konami_code:
                dialog = Message(self, "Someone wrote the KONAMI CODE !",
                    "Nice catch ! You have discover an easter-egg ! But, this "
                    "kind of code is usefull in a game, not in an emulators "
                    "manager !", Icons.Monkey)

                dialog.set_size(500, -1)

                dialog.run()
                dialog.destroy()

                self.keys = list()

            if not self.keys == konami_code[0:len(self.keys)]:
                self.keys = list()


    def set_informations(self):
        """ Update headerbar title and subtitle
        """

        game = self.selection["game"]
        console = self.selection["console"]

        self.set_informations_headerbar(game)

        # ----------------------------
        #   Game informations
        # ----------------------------

        if game is not None:

            if console is not None:
                self.separator_game.show()

                self.label_game_title.set_markup(
                    "<span weight='bold' size='x-large'>%s</span>" % \
                    replace_for_markup(game.name))

                # Get rom specified emulator
                emulator = console.emulator

                if game.emulator is not None:
                    emulator = game.emulator

                # ----------------------------
                #   Show screenshot
                # ----------------------------

                results = sorted(emulator.get_screenshots(game))

                # Check if rom has some screenshots
                if len(results) > 0:
                    index = -1

                    # Get a random file from rom screenshots
                    if self.config.getboolean(
                        "gem", "show_random_screenshot", fallback=True):
                        index = randint(0, len(results) - 1)

                    orientation = self.paned_games.get_orientation()

                    if orientation == Gtk.Orientation.HORIZONTAL:
                        pixbuf = Pixbuf.new_from_file_at_scale(
                            results[index], 400, -1, True)

                    else:
                        pixbuf = Pixbuf.new_from_file_at_scale(
                            results[index], -1, 200, True)

                    if pixbuf is not None:
                        self.image_game_screen.set_from_pixbuf(pixbuf)

                else:
                    self.image_game_screen.set_from_pixbuf(None)

                # ----------------------------
                #   Show informations
                # ----------------------------

                # Play time
                if not game.play_time == timedelta():
                    self.widgets_sidebar["play_time"]["box"].show_all()
                    self.widgets_sidebar["play_time"]["value"].set_markup(
                        string_from_time(game.play_time))

                else:
                    self.widgets_sidebar["play_time"]["box"].hide()
                    self.widgets_sidebar["play_time"]["value"].set_text(str())

                # Last launch
                if game.last_launch_date is not None:
                    self.widgets_sidebar["last_play"]["box"].show_all()
                    self.widgets_sidebar["last_play"]["value"].set_markup(
                        string_from_date(game.last_launch_date))
                else:
                    self.widgets_sidebar["last_play"]["box"].hide()
                    self.widgets_sidebar["last_play"]["value"].set_text(str())

                # Tags
                if len(game.tags) > 0:
                    self.widgets_sidebar["tags"]["box"].show_all()
                    self.widgets_sidebar["tags"]["value"].set_markup(
                        replace_for_markup(', '.join(game.tags)))
                else:
                    self.widgets_sidebar["tags"]["box"].hide()
                    self.widgets_sidebar["tags"]["value"].set_text(str())

                # Game emulator
                if emulator is not None:
                    self.separator_footer.show()

                    self.label_game_footer.set_markup("<b>%s</b> : %s" % (
                        _("Emulator"), emulator.name))

                    # Game savestates
                    if len(emulator.get_savestates(game)) > 0:
                        self.image_paned_savestates.set_from_pixbuf(
                            self.icons["save"])
                    else:
                        self.image_paned_savestates.set_from_pixbuf(
                            self.alternative["save"])

                    # Game custom parameters
                    if (game.emulator is not None and \
                        not game.emulator.name == console.emulator.name) or \
                        game.default is not None:
                        self.image_paned_parameters.set_from_pixbuf(
                            self.icons["except"])
                    else:
                        self.image_paned_parameters.set_from_pixbuf(
                            self.alternative["except"])

                else:
                    self.separator_footer.hide()

                    self.image_paned_savestates.set_from_pixbuf(self.empty)
                    self.image_paned_parameters.set_from_pixbuf(self.empty)

        else:
            for key, value in self.sidebar_keys:
                self.widgets_sidebar[key]["box"].hide()

            self.separator_game.hide()
            self.separator_footer.hide()

            self.label_game_title.set_text(str())
            self.label_game_footer.set_text(str())
            self.label_game_description.set_text(str())

            self.image_game_screen.set_from_pixbuf(None)

            self.image_paned_savestates.set_from_pixbuf(self.empty)
            self.image_paned_parameters.set_from_pixbuf(self.empty)


    def set_informations_headerbar(self, game):
        """ Update headerbar and statusbar informations from games list

        Parameters
        ----------
        game : gem.api.Game
            Game object
        """

        texts = list()

        if(len(self.filter_games) == 1):
            texts = [_("1 game available")]
        elif(len(self.filter_games) > 1):
            texts = [_("%s games availables") % len(self.filter_games)]

        if game is not None:
            texts.append(game.name)

        # ----------------------------
        #   Interface theme specific
        # ----------------------------

        # Default theme
        self.headerbar.set_subtitle(" - ".join(texts))

        # Classic theme
        self.statusbar.push(0, " - ".join(texts))


    def set_message(self, title, message, icon="dialog-error", popup=True):
        """ Open a message dialog

        This function open a dialog to inform user and write message to logger
        output.

        Parameters
        ----------
        title : str
            Dialog title
        message : str
            Dialog message
        icon : str
            Dialog icon, set also the logging mode
        """

        if icon == Icons.Error:
            self.logger.error(message)
        elif icon == Icons.Warning:
            self.logger.warning(message)
        else:
            self.logger.info(message)

        if popup:
            dialog = Message(self, title, message, icon)

            dialog.run()
            dialog.destroy()


    def set_infobar(self,
        text=str(), log=str(), message_type=Gtk.MessageType.INFO):
        """ Set infobar content

        This function set the infobar widget to inform user for specific things

        Others Parameters
        -----------------
        text : str
            Message text (Default: None)
        text : str
            Logger text (Default: None)
        message_type : Gtk.MessageType
            Message type (Default: Gtk.MessageType.INFO)
        """

        if len(log) == 0 and len(text) > 0:
            log = text

        # Set a logger message
        if len(log) > 0:
            if message_type is Gtk.MessageType.ERROR:
                self.logger.error(log)
            elif message_type is Gtk.MessageType.WARNING:
                self.logger.warning(log)

        self.infobar.set_message_type(message_type)

        # Set infobar visibility
        if len(text) > 0:
            if len(self.grid_infobar.get_children()) == 0:
                self.grid_infobar.pack_start(self.infobar, True, True, 0)

        elif len(self.grid_infobar.get_children()) > 0:
            self.grid_infobar.remove(self.infobar)

        self.label_infobar.set_markup(text)


    def __on_show_about(self, *args):
        """ Show about dialog
        """

        self.set_sensitive(False)

        about = Gtk.AboutDialog(use_header_bar=not self.use_classic_theme)

        about.set_transient_for(self)

        about.set_program_name(GEM.Name)
        about.set_version("%s (%s)" % (GEM.Version, GEM.CodeName))
        about.set_comments(GEM.Description)
        about.set_copyright(GEM.Copyleft)
        about.set_website(GEM.Website)

        about.set_logo_icon_name(GEM.Icon)

        about.set_authors([
            "Lubert Aurélien (PacMiam)" ])
        about.set_artists([
            "Tango projects - GPLv3",
            "Gelide projects - GPLv3",
            "Evan-Amos - CC-by-SA 3.0" ])
        about.set_translator_credits(
            _("translator-credits"))
        about.set_license_type(
            Gtk.License.GPL_3_0)

        # Strange case... With an headerbar, the AboutDialog got some useless
        # buttons whitout any reasons. To avoid this, I remove any widget from
        # headerbar which is not a Gtk.StackSwitcher.
        if not self.use_classic_theme:
            children = about.get_header_bar().get_children()

            for child in children:
                if type(child) is not Gtk.StackSwitcher:
                    about.get_header_bar().remove(child)

        about.run()

        self.set_sensitive(True)
        about.destroy()


    def __on_show_viewer(self, *args):
        """ Show game screenshots

        This function open game screenshots in a viewer. This viewer can be a
        custom one or the gem native viewer. This choice can be do in gem
        configuration file
        """

        game = self.selection["game"]
        console = self.selection["console"]

        if game is not None and console is not None:
            # Get name and extension from game
            gamename, ext = splitext(game.path[-1])

            # Get rom specified emulator
            emulator = console.emulator

            if game.emulator is not None:
                emulator = game.emulator

            # Check if rom has some screenshots
            results = emulator.get_screenshots(game)

            # ----------------------------
            #   Show screenshots viewer
            # ----------------------------

            if len(results) > 0:
                title = "%s (%s)" % (game.name, console.name)

                self.set_sensitive(False)

                # Get external viewer
                viewer = self.config.get("viewer", "binary")

                if self.config.getboolean("viewer", "native", fallback=True):
                    try:
                        size = self.config.get(
                            "windows", "viewer", fallback="800x600").split('x')

                    except ValueError as error:
                        size = (800, 600)

                    dialog = DialogViewer(self, title, size, sorted(results))
                    dialog.run()

                    self.config.modify(
                        "windows", "viewer", "%dx%d" % dialog.get_size())
                    self.config.update()

                    dialog.destroy()

                elif exists(viewer):
                    command = list()

                    # Append binaries
                    command.extend(shlex_split(viewer))

                    # Append arguments
                    args = self.config.item("viewer", "options")

                    if args is not None:
                        command.extend(shlex_split(args))

                    # Append game file
                    command.extend(sorted(results))

                    process = Popen(command)
                    process.wait()

                else:
                    self.set_message(_("Cannot open screenshots viewer"),
                        _("Cannot find <b>%s</b>") % viewer, Icons.Warning)

                self.set_sensitive(True)

                # ----------------------------
                #   Check screenshots
                # ----------------------------

                if len(emulator.get_screenshots(game)) == 0:
                    self.set_game_data(Columns.Snapshots,
                        self.alternative["snap"], game.filename)


    def __on_show_preferences(self, *args):
        """ Show preferences window

        This function show the gem preferences manager
        """

        self.set_sensitive(False)

        dialog = Preferences(self.api, self)

        if dialog.run() == Gtk.ResponseType.APPLY:
            dialog.save_configuration()

            self.logger.debug("Main interface need to be reload")
            self.load_interface()

        dialog.destroy()

        self.set_sensitive(True)


    def __on_show_log(self, *args):
        """ Show gem log

        This function show the gem log content in a non-editable dialog
        """

        path = path_join(GEM.Local, "gem.log")

        game = self.selection["game"]

        if path is not None and exists(expanduser(path)):
            try:
                size = self.config.get(
                    "windows", "log", fallback="800x600").split('x')

            except ValueError as error:
                size = (800, 600)

            self.set_sensitive(False)

            dialog = DialogEditor(
                self, _("GEM"), expanduser(path), size, False, Icons.Terminal)

            dialog.run()

            self.config.modify("windows", "log", "%dx%d" % dialog.get_size())
            self.config.update()

            self.set_sensitive(True)

            dialog.destroy()


    def __on_show_notes(self, *args):
        """ Edit game notes

        This function allow user to write some notes for a specific game. The
        user can open as many notes he wants but cannot open a note already
        open
        """

        game = self.selection["game"]

        if game is not None:
            path = path_join(GEM.Local, "notes", game.filename + ".txt")

            if path is not None and not expanduser(path) in self.notes.keys():
                try:
                    size = self.config.get(
                        "windows", "notes", fallback="800x600").split('x')

                except ValueError as error:
                    size = (800, 600)

                dialog = DialogEditor(self,
                    game.name, expanduser(path), size, icon=Icons.Document)

                # Allow to launch games with open notes
                dialog.set_modal(False)

                dialog.window.connect("response", self.__on_show_notes_response,
                    dialog, game.name, expanduser(path))

                dialog.show_all()

                # Save dialogs to close it properly when gem terminate and avoid
                # to reopen existing one
                self.notes[expanduser(path)] = dialog

            elif expanduser(path) in self.notes.keys():
                self.notes[expanduser(path)].grab_focus()


    def __on_show_notes_response(self, widget, response, dialog, title, path):
        """ Close notes dialog

        This function close current notes dialog and save his textview buffer to
        the game notes file

        Parameters
        ----------
        widget : Gtk.Dialog
            Dialog object
        response : Gtk.ResponseType
            Dialog object user response
        dialog : gem.windows.DialogEditor
            Dialog editor object
        title : str
            Dialog title, it's game name by default
        path : str
            Notes path
        """

        if response == Gtk.ResponseType.APPLY:
            text_buffer = dialog.buffer_editor.get_text(
                dialog.buffer_editor.get_start_iter(),
                dialog.buffer_editor.get_end_iter(), True)

            if len(text_buffer) > 0:
                with open(path, 'w') as pipe:
                    pipe.write(text_buffer)

                self.logger.info(_("Update note for %s") % title)

        self.config.modify("windows", "notes", "%dx%d" % dialog.get_size())
        self.config.update()

        dialog.destroy()

        if path in self.notes.keys():
            del self.notes[path]


    def __on_show_emulator_config(self, *args):
        """ Edit emulator configuration file
        """

        game = self.selection["game"]
        console = self.selection["console"]

        if console is not None:
            emulator = console.emulator

            if game is not None and game.emulator is not None:
                emulator = game.emulator

            if emulator.configuration is not None:
                path = emulator.configuration

                if path is not None and exists(expanduser(path)):
                    try:
                        size = self.config.get(
                            "windows", "editor", fallback="800x600").split('x')

                    except ValueError as error:
                        size = (800, 600)

                    self.set_sensitive(False)

                    dialog = DialogEditor(self, _("Edit %s configuration") % (
                        emulator.name), expanduser(path), size)

                    if dialog.run() == Gtk.ResponseType.APPLY:
                        with open(path, 'w') as pipe:
                            pipe.write(dialog.buffer_editor.get_text(
                                dialog.buffer_editor.get_start_iter(),
                                dialog.buffer_editor.get_end_iter(), True))

                        self.logger.info(
                            _("Update %s configuration file") % emulator.name)

                    self.config.modify(
                        "windows", "editor", "%dx%d" % dialog.get_size())
                    self.config.update()

                    self.set_sensitive(True)

                    dialog.destroy()


    def append_consoles(self):
        """ Append to consoles combobox all available consoles

        This function add every consoles into consoles combobox and inform user
        when an emulator binary is missing

        Returns
        -------
        Gtk.TreeIter or None
            Selected row
        """

        item = None

        size = 24

        # Reset consoles caches
        self.consoles_iter.clear()
        self.model_consoles.clear()

        # Get toolbar icons size
        data = Gtk.IconSize.lookup(self.toolbar.get_icon_size())
        if data is not None and len(data) == 3:
            # Get maximum icon size (this size cannot be under 24 pixels)
            size = max(size, data[1], data[2])

        for console in self.api.consoles:
            console = self.api.get_console(console)

            # Reload games list
            console.set_games(self.api)

            # Check if console ROM path exist
            if console.emulator is not None and exists(console.path):
                status = self.empty

                hide = self.config.getboolean(
                    "gem", "hide_empty_console", fallback=False)

                # Check if console ROM path is empty
                if not hide or (hide and len(console.games) > 0):

                    # Check if current emulator can be launched
                    if not console.emulator.exists:
                        status = self.icons["warning"]

                        self.logger.warning(
                            _("Cannot find %(binary)s for %(console)s") % {
                                "binary": console.emulator.binary,
                                "console": console.name })

                    # Get console icon
                    icon = icon_from_data(
                        console.icon, self.empty, size, size, "consoles")

                    # Append a new console in combobox model
                    row = self.model_consoles.append(
                        [icon, console.name, status, console.id])

                    # Store console iter
                    self.consoles_iter[console.id] = row

                    selection = self.selection.get("console")

                    if selection is not None and selection.name == console.name:
                        item = row

        if len(self.model_consoles) > 0:
            self.logger.debug(
                "%d console(s) has been added" % len(self.model_consoles))

        self.combo_consoles.set_wrap_width(int(len(self.model_consoles) / 10))

        return item


    def __on_selected_console(self, widget=None):
        """ Select a console

        This function occurs when the user select a console in the consoles
        combobox

        Other Parameters
        ----------------
        widget : Gtk.Widget
            Object which receive signal (Default: None)
        """

        self.selection["name"] = None
        self.selection["game"] = None

        self.set_infobar()
        self.set_informations()

        error = False

        treeiter = self.combo_consoles.get_active_iter()
        if treeiter is not None:

            console_id = self.model_consoles.get_value(treeiter, 3)
            if console_id is not None:
                console = self.api.get_console(console_id)

                # Save selected console
                self.selection["console"] = console

                # ------------------------------------
                #   Check data
                # ------------------------------------

                self.sensitive_interface()

                # Check console emulator
                if console.emulator is not None:
                    configuration = console.emulator.configuration

                    # Check emulator configurator
                    if configuration is not None and exists(configuration):
                        self.toolbar_item_properties.set_sensitive(True)
                    else:
                        self.toolbar_item_properties.set_sensitive(False)

                # Activate refresh button
                self.toolbar_item_refresh.set_sensitive(True)

                # ------------------------------------
                #   Load game list
                # ------------------------------------

                if not self.list_thread == 0:
                    source_remove(self.list_thread)

                loader = self.append_games(console)
                self.list_thread = idle_add(loader.__next__)

                self.selection["game"] = None


    def append_games(self, console):
        """ Append to games treeview all games from console

        This function add every games which match console extensions to games
        treeview

        Parameters
        ----------
        console : gem.api.Console
            Console object

        Raises
        ------
        TypeError
            if console type is not gem.api.Console

        Notes
        -----
        Using yield avoid an UI freeze when append a lot of games
        """

        if type(console) is not Console:
            raise TypeError("Wrong type for console, expected gem.api.Console")

        iteration = int()

        # Get current thread id
        current_thread_id = self.list_thread

        self.game_path = dict()

        # ------------------------------------
        #   Check errors
        # ------------------------------------

        if console.emulator is None:
            self.set_infobar(
                _("There is no default emulator set for this console"),
                _("Cannot find emulator for %s") % console.name,
                Gtk.MessageType.WARNING)

        elif not console.emulator.exists:
            self.set_infobar(_("<b>%s</b> cannot been found on your "
                "system") % console.emulator.name,
                _("%s emulator not exist") % console.emulator.name,
                Gtk.MessageType.ERROR)

        # ------------------------------------
        #   Load data
        # ------------------------------------

        self.scroll_games.set_visible(False)
        self.grid_games_placeholder.set_visible(True)

        self.model_games.clear()

        if console.emulator is not None:
            self.selection["console"] = console

            # ------------------------------------
            #   Refresh treeview
            # ------------------------------------

            self.treeview_games.set_enable_search(False)
            self.treeview_games.freeze_child_notify()

            # ------------------------------------
            #   Load games
            # ------------------------------------

            console.set_games(self.api)

            emulator = self.api.get_emulator(console.emulator.id)

            games = console.get_games()

            if len(games) > 0:
                self.scroll_games.set_visible(True)
                self.grid_games_placeholder.set_visible(False)

            yield True

            for game in games:

                # Another thread has been called by user, close this one
                if not current_thread_id == self.list_thread:
                    yield False

                # Hide games which match ignores regex
                show = True
                for element in console.ignores:
                    try:
                        if match(element, game.name, IGNORECASE) is not None:
                            show = False
                            break
                    except:
                        pass

                # Check if rom file exists
                if exists(game.filepath) and show:

                    row_data = [
                        game.favorite,
                        self.alternative["favorite"],
                        self.alternative["multiplayer"],
                        self.alternative["unfinish"],
                        game.name,
                        str(),          # Played
                        str(),          # Last launch date
                        str(),          # Last launch time
                        str(),          # Total play time
                        str(),          # Installed date
                        self.alternative["except"],
                        self.alternative["snap"],
                        self.alternative["save"],
                        game ]

                    # Favorite
                    if game.favorite:
                        row_data[Columns.Icon] = self.icons["favorite"]

                    # Multiplayer
                    if game.multiplayer:
                        row_data[Columns.Multiplayer] = \
                            self.icons["multiplayer"]

                    # Finish
                    if game.finish:
                        row_data[Columns.Finish] = \
                            self.icons["finish"]

                    # Played
                    if game.played > 0:
                        row_data[Columns.Played] = str(game.played)

                    # Last launch date
                    if game.last_launch_date is not None:
                        row_data[Columns.LastPlay] = \
                            string_from_date(game.last_launch_date)

                    # Last launch time
                    if not game.last_launch_time == timedelta():
                        row_data[Columns.LastTimePlay] = \
                            string_from_time(game.last_launch_time)

                    # Play time
                    if not game.play_time == timedelta():
                        row_data[Columns.TimePlay] = \
                            string_from_time(game.play_time)

                    # Parameters
                    if game.default is not None:
                        row_data[Columns.Except] = self.icons["except"]

                    elif game.emulator is not None:
                        if not game.emulator.name == console.emulator.name:
                            row_data[Columns.Except] = self.icons["except"]

                    # Installed time
                    row_data[Columns.Installed] = string_from_date(
                        datetime.fromtimestamp(getctime(game.filepath)).date())

                    # Get global emulator
                    rom_emulator = emulator

                    # Set specified emulator is available
                    if game.emulator is not None:
                        rom_emulator = game.emulator

                    # Snap
                    if len(rom_emulator.get_screenshots(game)) > 0:
                        row_data[Columns.Snapshots] = self.icons["snap"]

                    # Save state
                    if len(rom_emulator.get_savestates(game)) > 0:
                        row_data[Columns.Save] = self.icons["save"]

                    row = self.model_games.append(row_data)

                    self.game_path[game.filename] = [game, row]

                    iteration += 1
                    if (iteration % 20 == 0):
                        self.set_informations()

                        self.treeview_games.thaw_child_notify()
                        yield True
                        self.treeview_games.freeze_child_notify()

            # Restore options for packages treeviews
            self.treeview_games.set_enable_search(True)
            self.treeview_games.thaw_child_notify()

        self.set_informations()

        # ------------------------------------
        #   Cannot read games path
        # ------------------------------------

        if not access(console.path, W_OK):
            pass

        # ------------------------------------
        #   Close thread
        # ------------------------------------

        self.list_thread = int()

        yield False


    def __on_selected_game(self, treeview, event):
        """ Select a game

        This function occurs when the user select a game in the games treeview

        Parameters
        ----------
        treeview : Gtk.Treeview
            Object which receive signal
        event : Gdk.EventButton or Gdk.EventKey
            Event which triggered this signal
        """

        available_events = [
            EventType.BUTTON_PRESS,
            EventType._2BUTTON_PRESS,
            EventType._3BUTTON_PRESS
        ]

        game = None
        treeiter = None

        run_game = False

        # ----------------------------
        #   Keyboard
        # ----------------------------

        if event.type == EventType.KEY_RELEASE:
            model, treeiter = treeview.get_selection().get_selected()

        # ----------------------------
        #   Mouse
        # ----------------------------

        elif event.type in available_events and event.button in (1, 3):

            # Get selection from cursor position
            if event.type == EventType.BUTTON_PRESS:
                selection = treeview.get_path_at_pos(int(event.x), int(event.y))

                if selection is not None:
                    model = treeview.get_model()
                    treeiter = model.get_iter(selection[0])

            # Get selection from treeview
            else:
                model, treeiter = treeview.get_selection().get_selected()

            # Mouse - Double click with left mouse button
            if event.type == EventType._2BUTTON_PRESS and event.button == 1:
                run_game = True and not self.is_rename

        # ----------------------------
        #   Game selected
        # ----------------------------

        # Get game data
        if treeiter is not None:
            game = model.get_value(treeiter, Columns.Object)

        # Check if the selected game has already been showed
        same_game = (self.selection["game"] == game)

        self.selection["game"] = game

        if game is not None:
            console = self.selection["console"]

            if console is not None:
                if not same_game:
                    self.sensitive_interface(True)

                # Get Game emulator
                emulator = console.emulator
                if game.emulator is not None:
                    emulator = game.emulator

                # ----------------------------
                #   Game data
                # ----------------------------

                if game.filename in self.threads:
                    self.__on_game_launch_button_update(False)
                    self.headerbar_item_launch.set_sensitive(True)

                    self.menu_item_launch.set_sensitive(False)
                    self.menu_item_database.set_sensitive(False)
                    self.menu_item_remove.set_sensitive(False)
                    self.menu_item_mednafen.set_sensitive(False)

                    self.menubar_main_item_launch.set_sensitive(False)
                    self.menubar_edit_item_database.set_sensitive(False)
                    self.menubar_edit_item_delete.set_sensitive(False)
                    self.menubar_edit_item_mednafen.set_sensitive(False)

                if not same_game:
                    # Check extension and emulator for GBA game on mednafen
                    if not game.extension.lower() == ".gba" or \
                        not "mednafen" in emulator.binary or \
                        not self.__mednafen_status:
                        self.menu_item_mednafen.set_sensitive(False)
                        self.menubar_edit_item_mednafen.set_sensitive(False)

                    iter_snaps = model.get_value(treeiter, Columns.Snapshots)

                    # Check snaps icon to avoid to check screenshots again
                    if iter_snaps == self.alternative["snap"]:
                        self.toolbar_item_screenshots.set_sensitive(False)
                        self.menu_item_screenshots.set_sensitive(False)
                        self.menubar_main_item_screenshots.set_sensitive(False)

                    if self.check_log() is None:
                        self.toolbar_item_output.set_sensitive(False)
                        self.menu_item_output.set_sensitive(False)
                        self.menubar_main_item_output.set_sensitive(False)

                if run_game:
                    self.__on_game_launch()

        else:
            self.treeview_games.get_selection().unselect_all()

            self.selection["game"] = None

            self.__on_game_launch_button_update(True)
            self.headerbar_item_launch.set_sensitive(False)

        if not same_game:
            self.set_informations()


    def __on_selected_game_tooltip(self, treeview, x, y, keyboard, tooltip):
        """ Show game informations tooltip

        Parameters
        ----------
        treeview : Gtk.TreeView
            Widget which received the signal
        x : int
            X coordinate for mouse cursor position
        y : int
            Y coordinate for mouse cursor position
        keyboard : bool
            Set as True if the tooltip has been trigged by keyboard
        tooltip : Gtk.Tooltip
            Tooltip widget

        Returns
        -------
        bool
            Tooltip visible status
        """

        # Show a tooltip when the main window is sentitive only
        if self.get_sensitive():

            # Get relative treerow position based on absolute cursor coordinates
            x, y = treeview.convert_widget_to_bin_window_coords(x, y)

            selection = treeview.get_path_at_pos(x, y)
            if selection is not None:
                model = treeview.get_model()
                treeiter = model.get_iter(selection[0])

                game = model.get_value(treeiter, Columns.Object)

                # Reload tooltip when another game is hovered
                if not self.__current_tooltip == game:
                    self.__current_tooltip = game
                    self.__current_tooltip_data = list()
                    self.__current_tooltip_pixbuf = None

                    return False

                # Get new data from hovered game
                if len(self.__current_tooltip_data) == 0:
                    data = list()

                    data.append(
                        "<big><b>%s</b></big>" % replace_for_markup(game.name))

                    if not game.play_time == timedelta():
                        data.append(
                            ": ".join(
                                [
                                    "<b>%s</b>" % _("Play time"),
                                    parse_timedelta(game.play_time)
                                ]
                            )
                        )

                    if not game.last_launch_time == timedelta():
                        data.append(
                            ": ".join(
                                [
                                    "<b>%s</b>" % _("Last launch"),
                                    parse_timedelta(game.last_launch_time)
                                ]
                            )
                        )

                    # Fancy new line
                    if len(data) > 1:
                        data.insert(1, str())

                    self.__current_tooltip_data = data

                console = self.selection["console"]

                # Get new screenshots from hovered game
                if console is not None and \
                    self.__current_tooltip_pixbuf is None:

                    # Get Game emulator
                    emulator = console.emulator
                    if game.emulator is not None:
                        emulator = game.emulator

                    screenshots = sorted(emulator.get_screenshots(game))
                    if len(screenshots) > 0:
                        pixbuf = Pixbuf.new_from_file(screenshots[-1])

                        # Resize pixbuf to have a 96 pixels height
                        pixbuf = pixbuf.scale_simple(pixbuf.get_width() * float(
                            96 / pixbuf.get_height()), 96, InterpType.TILES)

                        self.__current_tooltip_pixbuf = pixbuf

                # Only show tooltip when data are available
                if len(self.__current_tooltip_data) > 0:
                    tooltip.set_markup('\n'.join(self.__current_tooltip_data))

                    if self.__current_tooltip_pixbuf is not None:
                        tooltip.set_icon(self.__current_tooltip_pixbuf)

                    self.__current_tooltip = game

                    return True

        return False


    def check_selection(self):
        """ Check selected game

        This function check if the selected game in the games treeview is the
        same as the one stock in self.selection["game"]. If this is not the
        case, the self.selection is reset

        Returns
        -------
        bool
            Check status
        """

        name = None

        if self.selection["game"] is not None:
            name = self.selection["game"].name

        if name is not None:
            model, treeiter = self.treeview_games.get_selection().get_selected()

            if treeiter is not None:
                treeview_name = model.get_value(treeiter, Columns.Name)

                if not treeview_name == name:
                    self.treeview_games.get_selection().unselect_iter(treeiter)
                    self.selection["game"] = None

                    self.sensitive_interface()
                    self.set_informations()

                    return False

        return True


    def __on_game_launch_button_update(self, status):
        """ Update game launch button

        Parameters
        ----------
        status : bool
            The game status
        """

        label = self.headerbar_item_launch.get_label()

        if status and not label == _("Play"):
            self.headerbar_item_launch.set_label(
                _("Play"))
            self.headerbar_item_launch.set_tooltip_text(
                _("Launch selected game"))
            self.headerbar_item_launch.get_style_context().remove_class(
                "destructive-action")
            self.headerbar_item_launch.get_style_context().add_class(
                "suggested-action")

        elif not status and not label == _("Stop"):
            self.headerbar_item_launch.set_label(
                _("Stop"))
            self.headerbar_item_launch.set_tooltip_text(
                _("Stop selected game"))
            self.headerbar_item_launch.get_style_context().remove_class(
                "suggested-action")
            self.headerbar_item_launch.get_style_context().add_class(
                "destructive-action")


    def __on_game_launch(self, widget=None, *args):
        """ Prepare the game launch

        This function prepare the game launch and start a thread when everything
        are done
        """

        binary = str()

        # ----------------------------
        #   Check selection
        # ----------------------------

        game = self.selection["game"]

        if game is None:
            return False

        if not self.check_selection():
            return False

        if game.filename in self.threads:
            if widget is not None and type(widget) is Gtk.Button:
                self.threads[game.filename].proc.terminate()

            return False

        if self.is_rename:
            return False

        # ----------------------------
        #   Check emulator
        # ----------------------------

        console = self.selection["console"]

        if console is not None:
            emulator = console.emulator

            if game.emulator is not None:
                emulator = game.emulator

            if emulator is not None and emulator.id in self.api.emulators:
                self.logger.info(_("Initialize %s") % game.name)

                # ----------------------------
                #   Generate correct command
                # ----------------------------

                try:
                    command = emulator.command(game,
                        self.headerbar_item_fullscreen.get_active())

                except FileNotFoundError as error:
                    self.set_message(_("Cannot launch game"),
                        _("%s binary cannot be found") % emulator.name)
                    return False

                if len(command) > 0:

                    # ----------------------------
                    #   Run game
                    # ----------------------------

                    thread = GameThread(self, console, emulator, game, command)

                    # Save thread references
                    self.threads[game.filename] = thread

                    # Launch thread
                    thread.start()

                    self.__on_game_launch_button_update(False)
                    self.headerbar_item_launch.set_sensitive(True)

                    self.menu_item_output.set_sensitive(True)
                    self.toolbar_item_output.set_sensitive(True)
                    self.menubar_main_item_output.set_sensitive(True)

                    return True

        return False


    def __on_game_terminate(self, widget, thread):
        """ Terminate the game processus and update data

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        thread : gem.game.GameThread
            Game thread
        """

        # ----------------------------
        #   Save game data
        # ----------------------------

        emulator = thread.emulator

        if not thread.error:

            # Get the last occurence from database
            game = self.api.get_game(thread.console.id, thread.game.id)

            # ----------------------------
            #   Update data
            # ----------------------------

            # Play data
            game.played += 1
            game.play_time = thread.game.play_time + thread.delta
            game.last_launch_time = thread.delta
            game.last_launch_date = date.today()

            # Update game from database
            self.api.update_game(game)

            # Played
            self.set_game_data(Columns.Played, str(game.played), game.filename)

            # Last played
            self.set_game_data(Columns.LastPlay,
                string_from_date(game.last_launch_date), game.filename)

            # Last time played
            self.set_game_data(Columns.LastTimePlay,
                string_from_time(game.last_launch_time), game.filename)

            # Play time
            self.set_game_data(Columns.TimePlay,
                string_from_time(game.play_time), game.filename)

            # Snaps
            if len(emulator.get_screenshots(game)) > 0:
                self.set_game_data(
                    Columns.Snapshots, self.icons["snap"], game.filename)
                self.toolbar_item_screenshots.set_sensitive(True)
                self.menubar_main_item_screenshots.set_sensitive(True)

            else:
                self.set_game_data(Columns.Snapshots,
                    self.alternative["snap"], game.filename)

            # Save state
            if len(emulator.get_savestates(game)) > 0:
                self.set_game_data(
                    Columns.Save, self.icons["save"], game.filename)

            else:
                self.set_game_data(
                    Columns.Save, self.alternative["save"], game.filename)

            self.set_informations()

        # ----------------------------
        #   Refresh widgets
        # ----------------------------

        # Get current selected file
        select_game = self.selection["game"]

        # Check if current selected file is the same as thread file
        if select_game is not None and select_game.id == game.id:
            self.logger.debug("Restore widgets status for %s" % game.name)

            self.__on_game_launch_button_update(True)
            self.headerbar_item_launch.set_sensitive(True)

            self.menu_item_launch.set_sensitive(True)
            self.menu_item_database.set_sensitive(True)
            self.menu_item_remove.set_sensitive(True)

            self.menubar_main_item_launch.set_sensitive(True)
            self.menubar_edit_item_database.set_sensitive(True)
            self.menubar_edit_item_delete.set_sensitive(True)

            # Check extension and emulator for GBA game on mednafen
            if not game.extension.lower() == ".gba" or \
                not "mednafen" in emulator.binary or \
                not self.__mednafen_status:
                self.menu_item_mednafen.set_sensitive(True)
                self.menubar_edit_item_mednafen.set_sensitive(True)

            # Avoid to launch the game again when use Enter in game terminate
            # self.treeview_games.get_selection().unselect_all()

        # Remove this game from threads list
        if game.filename in self.threads:
            self.logger.debug("Remove %s from process cache" % game.name)

            del self.threads[game.filename]

        if len(self.threads) == 0:
            self.menu_item_preferences.set_sensitive(True)
            self.menubar_tools_item_preferences.set_sensitive(True)


    def __on_activate_renamed(self, *args):
        """ Activate name edit mode in games treeview
        """

        game = self.selection["game"]

        if game is not None:
            treeiter = self.game_path[game.filename][1]

            # Set edit mode for selected treeiter
            self.treeview_games.set_cursor(self.model_games.get_path(treeiter),
                self.column_game_name, True)


    def __on_game_renamed_started(self, *args):
        """ Update is_rename status when started game renaming
        """

        self.is_rename = True


    def __on_game_renamed_cancel(self, *args):
        """ Update is_rename status when canceled game renaming
        """

        self.is_rename = False


    def __on_game_renamed(self, widget, path, new_name):
        """ Set a custom name for a specific game

        Parameters
        ----------
        widget : Gtk.CellRendererText
            object which received the signal
        path : str
            path identifying the edited cell
        new_name : str
            new name
        """

        game = self.model_games[path][Columns.Object]

        if game is not None:
            selected_game = self.selection["game"]

            # Avoid to validate rename when the user click outside
            if selected_game is not None and game.id == selected_game.id:

                # Check if game name has been changed
                if not new_name == game.name:
                    self.logger.info(_("Rename %(old)s to %(new)s") % {
                        "old": game.name, "new": new_name })

                    # Update game name
                    self.model_games[path][Columns.Name] = str(new_name)

                    game.name = new_name

                    # Update game from database
                    self.api.update_game(game)

                    # Store modified game
                    self.selection["game"] = game

                    self.set_informations()

        self.__on_game_renamed_cancel()


    def __on_game_clean(self, *args):
        """ Reset game informations from database
        """

        game = self.selection["game"]

        if game is not None:
            treeiter = self.game_path[game.filename][1]

            self.set_sensitive(False)

            dialog = Question(self, _("Reset game"),
                _("Would you really want to reset this game informations ?"))

            if dialog.run() == Gtk.ResponseType.YES:
                data = {
                    Columns.Name: game.filename,
                    Columns.Favorite: False,
                    Columns.Icon: self.alternative["favorite"],
                    Columns.Played: None,
                    Columns.LastPlay: None,
                    Columns.TimePlay: None,
                    Columns.LastTimePlay: None,
                    Columns.Except: self.alternative["except"],
                    Columns.Multiplayer: self.alternative["multiplayer"],
                }

                for key, value in data.items():
                    self.model_games[treeiter][key] = value

                game.reset()

                # Update game from database
                self.api.update_game(game)

                self.set_informations()

            self.set_sensitive(True)

            dialog.destroy()


    def __on_game_removed(self, *args):
        """ Remove a game

        This function also remove files from user disk as screenshots,
        savestates and game file.
        """

        game = self.selection["game"]
        console = self.selection["console"]

        if game is not None and console is not None:

            # Avoid trying to remove an executed game
            if not game.filename in self.threads:
                treeiter = self.game_path[game.filename][1]

                emulator = console.emulator

                file_to_remove = list()

                need_to_reload = False

                # ----------------------------
                #   Dialog
                # ----------------------------

                self.set_sensitive(False)

                dialog = DialogRemove(self, game)

                if dialog.run() == Gtk.ResponseType.YES:
                    file_to_remove.append(game.filepath)

                    # ----------------------------
                    #   Database
                    # ----------------------------

                    if dialog.check_database.get_active():
                        # Remove game from database
                        self.api.delete_game(game)

                    # ----------------------------
                    #   Emulator specific files
                    # ----------------------------

                    if emulator is not None:

                        # Savestates files
                        if dialog.check_save_state.get_active():
                            file_to_remove.extend(
                                emulator.get_savestates(game))

                        # Screenshots files
                        if dialog.check_screenshots.get_active():
                            file_to_remove.extend(
                                emulator.get_screenshots(game))

                    # ----------------------------
                    #   Remove files from disk
                    # ----------------------------

                    for element in file_to_remove:
                        self.logger.info(
                            _("%s has been deleted from disk") % element)

                        remove(element)

                    need_to_reload = True

                self.set_sensitive(True)

                dialog.destroy()

                if need_to_reload:
                    self.load_interface()

                    self.set_message(_("Remove a game"),
                        _("This game was removed successfully"),
                        Icons.Information)


    def __on_game_parameters(self, *args):
        """ Manage game default parameters

        This function allow the user to specify default emulator and default
        emulator arguments for the selected game
        """

        game = self.selection["game"]

        if game is not None:
            console = self.selection["console"]

            parameters = None

            emulator = {
                "rom": None,
                "console": None,
                "parameters": None }

            # Current console default emulator
            if console is not None and console.emulator is not None:
                emulator["console"] = console.emulator

                if console.emulator.default is not None:
                    parameters = console.emulator.default

            # ----------------------------
            #   Generate data
            # ----------------------------

            if game.emulator is not None and \
                not game.emulator == emulator["console"]:
                emulator["rom"] = game.emulator

            if game.default is not None and \
                not game.default == parameters:
                emulator["parameters"] = game.default

            # ----------------------------
            #   Dialog
            # ----------------------------

            self.set_sensitive(False)

            dialog = DialogParameters(self, game, emulator)

            if dialog.run() == Gtk.ResponseType.APPLY:
                self.logger.info(
                    _("Update %s parameters") % game.name)

                game.emulator = self.api.get_emulator(
                    dialog.combo.get_active_id())

                game.default = dialog.entry_arguments.get_text()
                if len(game.default) == 0:
                    game.default = None

                game.key = dialog.entry_key.get_text()
                if len(game.key) == 0:
                    game.key = None

                game.tags = dialog.entry_tags.get_text().split()

                game.environment = dict()

                for row in dialog.store_environment:
                    key = dialog.store_environment.get_value(row.iter, 0)

                    if key is not None and len(key) > 0:
                        value = dialog.store_environment.get_value(row.iter, 1)

                        if value is not None and len(value) > 0:
                            game.environment[key] = value

                # Update game from database
                self.api.update_game(game)

                # ----------------------------
                #   Check diferences
                # ----------------------------

                custom = False

                if game.emulator is not None and \
                    not game.emulator.name == console.emulator.name:
                    custom = True

                elif game.default is not None:
                    custom = True

                if custom:
                    self.set_game_data(Columns.Except,
                        self.icons["except"], game.filename)
                else:
                    self.set_game_data(Columns.Except,
                        self.alternative["except"], game.filename)

                # ----------------------------
                #   Update icons
                # ----------------------------

                new_emulator = emulator["console"]
                if game.emulator is not None:
                    new_emulator = game.emulator

                # Screenshots
                if len(new_emulator.get_screenshots(game)) > 0:
                    self.set_game_data(Columns.Snapshots,
                        self.icons["snap"], game.filename)

                    self.toolbar_item_screenshots.set_sensitive(True)
                    self.menubar_main_item_screenshots.set_sensitive(True)

                else:
                    self.set_game_data(Columns.Snapshots,
                        self.alternative["snap"], game.filename)

                    self.toolbar_item_screenshots.set_sensitive(False)
                    self.menubar_main_item_screenshots.set_sensitive(False)

                # Savestates
                if len(new_emulator.get_savestates(game)) > 0:
                    self.set_game_data(Columns.Save,
                        self.icons["save"], game.filename)

                else:
                    self.set_game_data(Columns.Save,
                        self.alternative["save"], game.filename)

                self.set_informations()

            self.set_sensitive(True)

            dialog.destroy()


    def __on_game_log(self, *args):
        """ Show game log

        This function show the gem log content in a non-editable dialog
        """

        path = self.check_log()

        game = self.selection["game"]

        if path is not None and exists(expanduser(path)):
            try:
                size = self.config.get(
                    "windows", "log", fallback="800x600").split('x')

            except ValueError as error:
                size = (800, 600)

            self.set_sensitive(False)

            dialog = DialogEditor(
                self, game.name, expanduser(path), size, False, Icons.Terminal)

            dialog.run()

            self.config.modify("windows", "log", "%dx%d" % dialog.get_size())
            self.config.update()

            self.set_sensitive(True)

            dialog.destroy()


    def __on_game_backup_memory(self, *args):
        """ Manage game backup memory

        This function can only be used with a GBA game and Mednafen emulator.
        """

        if self.menu_item_mednafen.get_sensitive():

            game = self.selection["game"]
            console = self.selection["console"]

            if game is not None and console is not None:
                content = dict()

                emulator = console.emulator
                if game.emulator is not None:
                    emulator = game.emulator

                # FIXME: Maybe a better way to determine type file
                filepath = expanduser(
                    path_join('~', ".mednafen", "sav", game.filename + ".type"))

                # Check if a type file already exist in mednafen sav folder
                if exists(filepath):
                    with open(filepath, 'r') as pipe:
                        for line in pipe.readlines():
                            data = line.split()

                            if len(data) == 2:
                                content[data[0]] = int(data[1])

                # ----------------------------
                #   Dialog
                # ----------------------------

                self.set_sensitive(False)

                dialog = DialogMednafenMemory(self, game.name, content)

                if dialog.run() == Gtk.ResponseType.APPLY:
                    data = list()
                    for key, value in dialog.model:
                        data.append(' '.join([key, str(value)]))

                    # Write data into type file
                    if len(data) > 0:
                        with open(filepath, 'w') as pipe:
                            pipe.write('\n'.join(data))

                    # Remove type file when no data are available
                    elif exists(filepath):
                        remove(filepath)

                self.set_sensitive(True)

                dialog.destroy()


    def __on_game_marked_as_favorite(self, *args):
        """ Mark or unmark a game as favorite

        This function update the database when user change the game favorite
        status
        """

        game = self.selection["game"]

        if game is not None:
            treeiter = self.game_path[game.filename][1]

            if not game.favorite:
                self.model_games[treeiter][Columns.Favorite] = True
                self.model_games[treeiter][Columns.Icon] = \
                    self.icons["favorite"]

                game.favorite = True

                # Update game from database
                self.api.update_game(game)

                self.logger.debug("Mark %s as favorite" % game.name)

            else:
                self.model_games[treeiter][Columns.Favorite] = False
                self.model_games[treeiter][Columns.Icon] = \
                    self.alternative["favorite"]

                game.favorite = False

                # Update game from database
                self.api.update_game(game)

                self.logger.debug("Unmark %s as favorite" % game.name)

            self.check_selection()


    def __on_game_marked_as_multiplayer(self, *args):
        """ Mark or unmark a game as multiplayer

        This function update the database when user change the game multiplayers
        status
        """

        game = self.selection["game"]

        if game is not None:
            treeiter = self.game_path[game.filename][1]

            if not game.multiplayer:
                self.model_games[treeiter][Columns.Multiplayer] = \
                    self.icons["multiplayer"]

                game.multiplayer = True

                # Update game from database
                self.api.update_game(game)

                self.logger.debug("Mark %s as multiplayers" % game.name)

            else:
                self.model_games[treeiter][Columns.Multiplayer] = \
                    self.alternative["multiplayer"]

                game.multiplayer = False

                # Update game from database
                self.api.update_game(game)

                self.logger.debug("Unmark %s as multiplayers" % game.name)

            self.check_selection()


    def __on_game_marked_as_finish(self, *args):
        """ Mark or unmark a game as finish

        This function update the database when user change the game finish
        status
        """

        game = self.selection["game"]

        if game is not None:
            treeiter = self.game_path[game.filename][1]

            if not game.finish:
                self.model_games[treeiter][Columns.Finish] = \
                    self.icons["finish"]

                game.finish = True

                # Update game from database
                self.api.update_game(game)

                self.logger.debug("Mark %s as finish" % game.name)

            else:
                self.model_games[treeiter][Columns.Finish] = \
                    self.alternative["finish"]

                game.finish = False

                # Update game from database
                self.api.update_game(game)

                self.logger.debug("Unmark %s as finish" % game.name)

            self.check_selection()


    def __on_game_copy(self, *args):
        """ Copy path folder which contains selected game to clipboard
        """

        game = self.selection["game"]

        if game is not None:
            self.clipboard.set_text(game.filepath, -1)


    def __on_game_open(self, *args):
        """ Open game directory

        This function open the folder which contains game with the default
        files manager

        Based on http://stackoverflow.com/a/6631329
        """

        game = self.selection["game"]

        if game is not None:
            path = game.path[0]

            self.logger.debug("Open %s folder in files manager" % path)

            if system() == "Windows":
                from os import startfile
                startfile(path)

            elif system() == "Darwin":
                Popen(["open", path], stdout=PIPE, stdin=PIPE,
                    stderr=STDOUT, universal_newlines=True)

            else:
                Popen(["xdg-open", path], stdout=PIPE, stdin=PIPE,
                    stderr=STDOUT, universal_newlines=True)


    def __on_game_generate_desktop(self, *args):
        """ Generate application desktop file

        This function generate a .desktop file to allow user to launch the game
        from his favorite applications launcher
        """

        model, treeiter = self.treeview_games.get_selection().get_selected()

        if treeiter is not None:
            game = self.selection["game"]
            console = self.selection["console"]

            # ----------------------------
            #   Check emulator
            # ----------------------------

            emulator = console.emulator

            if game is not None and game.emulator is not None:
                emulator = game.emulator

            if emulator is not None and emulator.id in self.api.emulators:
                name = "%s.desktop" % game.filename

                # ----------------------------
                #   Fill template
                # ----------------------------

                icon = console.icon
                if not exists(icon):
                    icon = path_join(GEM.Local, "icons", "consoles",
                        '.'.join([icon, Icons.Ext]))

                values = {
                    "%name%": game.name,
                    "%icon%": icon,
                    "%path%": game.path[0],
                    "%command%": ' '.join(emulator.command(game)) }

                # Quote game path
                values["%command%"] = values["%command%"].replace(
                    game.filepath, "\"%s\"" % game.filepath)

                self.set_sensitive(False)

                try:
                    desktop = path_join("config", Documents.Desktop)

                    with open(get_data(desktop), 'r') as pipe:
                        template = pipe.readlines()

                    content = str()

                    # Replace custom variables
                    for line in template:
                        for key in values.keys():
                            line = line.replace(key, values[key])

                        content += line

                    # Check ~/.local/share/applications
                    if not exists(Folders.Apps):
                        mkdir(Folders.Apps)

                    # Write the new desktop file
                    with open(path_join(Folders.Apps, name), 'w') as pipe:
                        pipe.write(content)

                    self.set_message(
                        _("Generate menu entry for %s") % game.name,
                        _("%s was generated successfully")  % name,
                        Icons.Information)

                except OSError as error:
                    self.set_message(
                        _("Generate menu entry for %s") % game.name,
                        _("An error occur during generation, consult log for "
                        "futher details."), Icons.Error)

                self.set_sensitive(True)


    def __on_menu_show(self, treeview, event):
        """ Open context menu

        This function open context-menu when user right-click or use context key
        on games treeview

        Parameters
        ----------
        treeview : Gtk.Treeview
            Object which receive signal
        event : Gdk.EventButton or Gdk.EventKey
            Event which triggered this signal

        Returns
        -------
        bool
            Context menu popup status
        """

        # Gdk.EventButton - Mouse
        if event.type == EventType.BUTTON_PRESS:
            if event.button == Gdk.BUTTON_SECONDARY:
                treeiter = treeview.get_path_at_pos(int(event.x), int(event.y))

                if treeiter is not None:
                    path, col = treeiter[0:2]

                    treeview.grab_focus()
                    treeview.set_cursor(path, col, 0)

                    self.menu_games.popup(
                        None, None, None, None, event.button, event.time)

                    return True

        # Gdk.EventKey - Keyboard
        elif event.type == EventType.KEY_RELEASE:
            if event.keyval == Gdk.KEY_Menu:
                model, treeiter = treeview.get_selection().get_selected()

                if treeiter is not None:
                    self.menu_games.popup(None, None, None, None, 0, event.time)

                    return True

        return False


    def __on_activate_fullscreen(self, widget, *args):
        """ Update fullscreen button

        This function alternate fullscreen status between active and inactive
        state when user use fullscreen button in toolbar

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        # Block signal
        self.menubar_main_item_fullscreen.handler_block(
            self.fullscreen_signal)
        self.headerbar_item_fullscreen.handler_block(
            self.fullscreen_signal_tool)

        # Switch fullscreen status
        self.__fullscreen_status = not self.__fullscreen_status

        if not self.__fullscreen_status:
            self.logger.debug("Switch game launch to windowed mode")

            self.headerbar_image_fullscreen.set_from_icon_name(
                Icons.Symbolic.Restore, Gtk.IconSize.SMALL_TOOLBAR)
            self.headerbar_item_fullscreen.get_style_context().remove_class(
                "suggested-action")

        else:
            self.logger.debug("Switch game launch to fullscreen mode")

            self.headerbar_image_fullscreen.set_from_icon_name(
                Icons.Symbolic.Fullscreen, Gtk.IconSize.SMALL_TOOLBAR)
            self.headerbar_item_fullscreen.get_style_context().add_class(
                "suggested-action")

        self.headerbar_item_fullscreen.set_active(self.__fullscreen_status)
        self.menubar_main_item_fullscreen.set_active(self.__fullscreen_status)

        # Unblock signal
        self.menubar_main_item_fullscreen.handler_unblock(
            self.fullscreen_signal)
        self.headerbar_item_fullscreen.handler_unblock(
            self.fullscreen_signal_tool)


    def __on_activate_dark_theme(self, widget, status=False, *args):
        """ Update dark theme status

        This function alternate between dark and light theme when user use
        dark theme entry in main menu

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal

        Others parameters
        -----------------
        status : bool
            New switch status (Default: False)
        """

        # Block signal to avoid stack overflow when toggled
        self.menu_item_dark_theme.handler_block(
            self.dark_signal_menu)
        self.menubar_tools_item_dark_theme.handler_block(
            self.dark_signal_menubar)

        dark_theme_status = not self.config.getboolean(
            "gem", "dark_theme", fallback=False)

        on_change_theme(dark_theme_status)

        self.config.modify("gem", "dark_theme", int(dark_theme_status))
        self.config.update()

        self.menu_item_dark_theme.set_active(dark_theme_status)
        self.menubar_tools_item_dark_theme.set_active(dark_theme_status)

        # Unblock signal
        self.menu_item_dark_theme.handler_unblock(
            self.dark_signal_menu)
        self.menubar_tools_item_dark_theme.handler_unblock(
            self.dark_signal_menubar)


    def __on_activate_sidebar(self, widget, status=False, *args):
        """ Update sidebar status

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal

        Others parameters
        -----------------
        status : bool
            New switch status (Default: False)
        """

        # Block signal to avoid stack overflow when toggled
        self.menu_item_sidebar.handler_block(
            self.side_signal_menu)
        self.menubar_tools_item_sidebar.handler_block(
            self.side_signal_menubar)

        sidebar_status = not self.config.getboolean(
            "gem", "show_sidebar", fallback=True)

        if sidebar_status:
            self.grid_paned.show()
        else:
            self.grid_paned.hide()

        self.config.modify("gem", "show_sidebar", int(sidebar_status))
        self.config.update()

        self.menu_item_sidebar.set_active(sidebar_status)
        self.menubar_tools_item_sidebar.set_active(sidebar_status)

        # Unblock signal
        self.menu_item_sidebar.handler_unblock(
            self.side_signal_menu)
        self.menubar_tools_item_sidebar.handler_unblock(
            self.side_signal_menubar)


    def __on_dnd_send_data(self, widget, context, data, info, time):
        """ Set rom file path uri

        This function send rom file path uri when user drag a game from gem and
        drop it to extern application

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        context : Gdk.DragContext
            Drag context
        data : Gtk.SelectionData
            Received data
        info : int
            info that has been registered with the target in the Gtk.TargetList
        time : int
            timestamp at which the data was received
        """

        if self.selection["game"] is not None:
            data.set_uris(["file://%s" % self.selection["game"].filepath])


    def __on_dnd_received_data(self, widget, context, x, y, data, info, time):
        """ Manage drag & drop acquisition

        This function receive drag files and install them into the correct
        games folder. If a file extension can be find in multiple consoles,
        a dialog appear to ask user where he want to put it.

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        context : Gdk.DragContext
            Drag context
        x : int
            X coordinate where the drop happened
        y : int
            Y coordinate where the drop happened
        data : Gtk.SelectionData
            Received data
        info : int
            info that has been registered with the target in the Gtk.TargetList
        time : int
            timestamp at which the data was received
        """

        self.logger.debug("Received data from drag & drop")

        widget.stop_emission("drag_data_received")

        # Current acquisition not respect text/uri-list
        if not info == 1337:
            return

        previous_console = None

        keep_console = False
        need_to_reload = False

        consoles_to_reload = dict()

        for uri in data.get_uris():
            result = urlparse(uri)

            console = None

            if result.scheme == "file":
                path = expanduser(url2pathname(result.path))

                if exists(path):
                    self.logger.debug("Check %s" % path)
                    filename, ext = splitext(basename(path))

                    # Lowercase extension
                    ext = ext.lower()

                    # ----------------------------
                    #   Get right console for rom
                    # ----------------------------

                    if not keep_console:

                        consoles_list = list()
                        for console in self.api.consoles:
                            console = self.api.get_console(console)

                            if console is not None:
                                extensions = console.extensions

                                if extensions is not None and \
                                    ext[1:] in extensions:
                                    consoles_list.append(console)

                        console = None

                        if len(consoles_list) > 0:
                            console = consoles_list[0]

                            if len(consoles_list) > 1:
                                self.set_sensitive(False)

                                dialog = DialogConsoles(self, basename(path),
                                    consoles_list, previous_console)

                                if dialog.run() == Gtk.ResponseType.APPLY:
                                    console = self.api.get_console(
                                        dialog.current)

                                    keep_console = dialog.switch.get_active()

                                    previous_console = console

                                else:
                                    console = None

                                dialog.destroy()

                                self.set_sensitive(True)

            # ----------------------------
            #   Check console
            # ----------------------------

            if console is not None:
                rom_path = expanduser(console.path)

                # ----------------------------
                #   Install roms
                # ----------------------------

                if rom_path is not None and not dirname(path) == rom_path and \
                    exists(rom_path) and access(rom_path, W_OK):
                    move = True

                    # Check if this game already exists in roms folder
                    if exists(path_join(rom_path, basename(path))):
                        dialog = Question(self, basename(path),
                            _("This rom already exists in %s. Do you want to "
                            "replace it ?") % rom_path)

                        move = False
                        if dialog.run() == Gtk.ResponseType.YES:
                            move = True

                            self.logger.debug(
                                "Move %s to %s" % (path, rom_path))

                            remove(path_join(rom_path, basename(path)))

                        dialog.destroy()

                    # The game can be moved in roms folder
                    if move:
                        rename(path, rom_path)

                        self.logger.info(_("Drop %(rom)s to %(path)s") % {
                            "rom": basename(path), "path": rom_path })

                        if console == self.selection["console"]:
                            need_to_reload = True

                        if not console.id in consoles_to_reload:
                            consoles_to_reload[console.id] = console

                        hide = self.config.getboolean(
                            "gem", "hide_empty_console", fallback=False)

                        # Console path is not empty
                        if hide and len(glob(path_join(rom_path, '*'))) == 1:
                            need_to_reload = True

                # ----------------------------
                #   Errors
                # ----------------------------

                if dirname(path) == rom_path:
                    pass

                elif not exists(rom_path):
                    self.set_message(basename(path), _("Destination %s not "
                        "exist. Canceling operation.") % rom_path)

                elif not access(rom_path, W_OK):
                    self.set_message(basename(path), _("Cannot write into %s. "
                        "Canceling operation.") % rom_path)

        # Reload console games
        for console in consoles_to_reload.values():
            console.set_games(self.api)

        # Reload interface when games list was modified
        if need_to_reload:
            self.load_interface()


    def check_desktop(self, filename):
        """ Check user applications folder for specific desktop file

        Parameters
        ----------
        filename : str
            Application name

        Returns
        -------
        bool
            Desktop file status

        Examples
        --------
        >>> check_desktop("unavailable_file")
        False

        Notes
        -----
        In GNU/Linux desktop, the default folder for user applications is:

            ~/.local/share/applications/
        """

        name, extension = splitext(filename)

        return exists(path_join(Folders.Apps, "%s.desktop" % name))


    def check_log(self):
        """ Check if a game has an output file available

        Returns
        -------
        str or None
            Output file path
        """

        game = self.selection["game"]

        if game is not None:
            log_path = path_join(GEM.Local, game.log)

            if exists(expanduser(log_path)):
                return log_path

        return None


    def check_mednafen(self):
        """ Check if Mednafen exists on user system

        This function read the first line of mednafen default output and check
        if this one match "Starting Mednafen [\d+\.?]+".

        Returns
        -------
        bool
            Mednafen exists status

        Notes
        -----
        Still possible to troll this function with a script call mednafen which
        send the match string as output. But, this problem only appear if a user
        want to do that, so ...
        """

        if len(get_binary_path("mednafen")) > 0:
            proc = Popen(
                [ "mednafen" ],
                stdin=PIPE,
                stdout=PIPE,
                stderr=STDOUT,
                universal_newlines=True)

            output, error_output = proc.communicate()

            if output is not None:
                result = match(
                    "Starting Mednafen [\d+\.?]+", output.split('\n')[0])

                if result is not None:
                    return True

        return False


    def set_game_data(self, index, data, gamename):
        """ Update game informations in games treeview

        Parameters
        ----------
        index : int
            Column index
        data : object
            Value to set
        gamename : str
            Game basename without extension
        """

        treeiter = self.game_path.get(gamename, None)

        if treeiter is not None:
            self.model_games[treeiter[1]][index] = data


    def emit(self, *args):
        """ Override emit function

        This override allow to use Interface function from another thread in
        MainThread
        """

        idle_add(GObject.emit, self, *args)


class Splash(Gtk.Window):

    def __init__(self, api):
        """ Constructor

        Parameters
        ----------
        api : gem.api.GEM
            GEM API instance

        Raises
        ------
        TypeError
            if api type is not gem.api.GEM
        """

        if not type(api) is GEM:
            raise TypeError("Wrong type for api, expected gem.api.GEM")

        Gtk.Window.__init__(self)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        # Set rows number for progressbar
        self.length = int()

        self.main_loop = None

        # ------------------------------------
        #   Initialize API
        # ------------------------------------

        # GEM API
        self.api = api

        # Quick access to API logger
        self.logger = api.logger

        # ------------------------------------
        #   Initialize icons
        # ------------------------------------

        # Get user icon theme
        self.icons_theme = Gtk.IconTheme.get_default()

        self.icons_theme.append_search_path(get_data(path_join("icons", "ui")))

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init packing
        self.__init_packing()

        # Start interface
        self.__start_interface()


    def __init_widgets (self):
        """ Load widgets into main interface
        """

        # ------------------------------------
        #   Main window
        # ------------------------------------

        self.set_title("Graphical Emulators Manager")

        self.set_modal(True)
        self.set_can_focus(True)
        self.set_resizable(False)
        self.set_keep_above(True)
        self.set_skip_taskbar_hint(True)
        self.set_type_hint(Gdk.WindowTypeHint.SPLASHSCREEN)

        self.set_position(Gtk.WindowPosition.CENTER)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.grid = Gtk.Box()

        # Properties
        self.grid.set_spacing(4)
        self.grid.set_border_width(16)
        self.grid.set_homogeneous(False)
        self.grid.set_orientation(Gtk.Orientation.VERTICAL)

        # ------------------------------------
        #   Logo
        # ------------------------------------

        self.label_splash = Gtk.Label()
        self.image_splash = Gtk.Image()

        # Properties
        self.label_splash.set_line_wrap(True)
        self.label_splash.set_use_markup(True)
        self.label_splash.set_line_wrap_mode(Pango.WrapMode.WORD)
        self.label_splash.set_markup(
            "<span weight='bold' size='x-large'>%s - %s</span>\n<i>%s</i>" % (
            GEM.Name, GEM.Version, GEM.CodeName))

        self.image_splash.set_from_icon_name(GEM.Icon, Gtk.IconSize.DND)
        self.image_splash.set_pixel_size(256)

        # ------------------------------------
        #   Progressbar
        # ------------------------------------

        self.label_progress = Gtk.Label()

        self.progressbar = Gtk.ProgressBar()

        # Properties
        self.label_progress.set_text(_("Migrating entries from old database"))
        self.label_progress.set_line_wrap_mode(Pango.WrapMode.WORD)
        self.label_progress.set_line_wrap(True)

        self.progressbar.set_show_text(True)


    def __init_packing(self):
        """ Initialize widgets packing in main window
        """

        self.grid.pack_start(self.image_splash, True, True, 0)
        self.grid.pack_start(self.label_splash, False, False, 8)
        self.grid.pack_start(self.label_progress, False, False, 8)
        self.grid.pack_start(self.progressbar, False, False, 0)

        self.add(self.grid)


    def __start_interface(self):
        """ Load data and start interface
        """

        self.show_all()

        self.label_progress.hide()
        self.progressbar.hide()

        self.refresh()

        # ------------------------------------
        #   Check migration
        # ------------------------------------

        idle_add(self.api.check_database, self)

        # ------------------------------------
        #   Main loop
        # ------------------------------------

        self.main_loop = MainLoop()
        self.main_loop.run()


    def close(self):
        """ Stop interface
        """

        # Sleep to avoid ultra quick splash
        sleep(0.42)

        # Rare case where the mainloop is not init when close() is running
        if self.main_loop is not None:
            self.main_loop.quit()

        self.destroy()


    def init(self, length):
        """ Initialize progressbar
        """

        self.length = length

        self.label_progress.show()
        self.progressbar.show()


    def update(self, index):
        """ Update progress in progressbar widgets
        """

        self.refresh()

        if index <= self.length:
            self.progressbar.set_text("%d / %d" % (index, self.length))
            self.progressbar.set_fraction(float(index) / (self.length))

            self.refresh()


    def refresh(self):
        """ Refresh all pendings event in main interface
        """

        while Gtk.events_pending():
            Gtk.main_iteration()
